#!/usr/bin/python3
# ----------------------------------------------------------------------
#    Copyright (C) 2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import unittest
from collections import namedtuple

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.logparser import ReadLog
from apparmor.rule.change_profile import ChangeProfileRule, ChangeProfileRuleset
from apparmor.translations import init_translation
from common_test import AATest, setup_all_loops

_ = init_translation()

exp = namedtuple(
    'exp', ('audit', 'allow_keyword', 'deny', 'comment',
            'execmode', 'execcond', 'all_execconds', 'targetprofile', 'all_targetprofiles'))

# --- tests for single ChangeProfileRule --- #


class ChangeProfileTest(AATest):
    def _compare_obj(self, obj, expected):
        self.assertEqual(expected.allow_keyword, obj.allow_keyword)
        self.assertEqual(expected.audit, obj.audit)
        self.assertEqual(expected.execmode, obj.execmode)
        self.assertEqual(expected.execcond, obj.execcond)
        self.assertEqual(expected.targetprofile, obj.targetprofile)
        self.assertEqual(expected.all_execconds, obj.all_execconds)
        self.assertEqual(expected.all_targetprofiles, obj.all_targetprofiles)
        self.assertEqual(expected.deny, obj.deny)
        self.assertEqual(expected.comment, obj.comment)


class ChangeProfileTestParse(ChangeProfileTest):
    tests = (
        # rawrule                                           audit  allow  deny   comment       execmode  execcond  all?   targetprof  all?
        ('change_profile,',                             exp(False, False, False, '',           None,     None,     True,  None,       True)),
        ('change_profile /foo,',                        exp(False, False, False, '',           None,     '/foo',   False, None,       True)),
        ('change_profile safe /foo,',                   exp(False, False, False, '',           'safe',   '/foo',   False, None,       True)),
        ('change_profile unsafe /foo,',                 exp(False, False, False, '',           'unsafe', '/foo',   False, None,       True)),
        ('change_profile /foo -> /bar,',                exp(False, False, False, '',           None,     '/foo',   False, '/bar',     False)),
        ('change_profile safe /foo -> /bar,',           exp(False, False, False, '',           'safe',   '/foo',   False, '/bar',     False)),
        ('change_profile unsafe /foo -> /bar,',         exp(False, False, False, '',           'unsafe', '/foo',   False, '/bar',     False)),
        ('deny change_profile /foo -> /bar, # comment', exp(False, False, True,  ' # comment', None,     '/foo',   False, '/bar',     False)),
        ('audit allow change_profile safe /foo,',       exp(True,  True,  False, '',           'safe',   '/foo',   False, None,       True)),
        ('change_profile -> /bar,',                     exp(False, False, False, '',           None,     None,     True,  '/bar',     False)),
        ('audit allow change_profile -> /bar,',         exp(True,  True,  False, '',           None,     None,     True,  '/bar',     False)),
        # quoted versions
        ('change_profile "/foo",',                      exp(False, False, False, '',           None,     '/foo',   False, None,       True)),
        ('change_profile "/foo" -> "/bar",',            exp(False, False, False, '',           None,     '/foo',   False, '/bar',     False)),
        ('deny change_profile "/foo" -> "/bar", # cmt', exp(False, False, True, ' # cmt',      None,     '/foo',   False, '/bar',     False)),
        ('audit allow change_profile "/foo",',          exp(True,  True,  False, '',           None,     '/foo',   False, None,       True)),
        ('change_profile -> "/bar",',                   exp(False, False, False, '',           None,     None,     True,  '/bar',     False)),
        ('audit allow change_profile -> "/bar",',       exp(True,  True,  False, '',           None,     None,     True,  '/bar',     False)),
        # with globbing and/or named profiles
        ('change_profile,',                             exp(False, False, False, '',           None,     None,     True,  None,       True)),
        ('change_profile /*,',                          exp(False, False, False, '',           None,     '/*',     False, None,       True)),
        ('change_profile /* -> bar,',                   exp(False, False, False, '',           None,     '/*',     False, 'bar',      False)),
        ('deny change_profile /** -> bar, # comment',   exp(False, False, True,  ' # comment', None,     '/**',    False, 'bar',      False)),
        ('audit allow change_profile /**,',             exp(True,  True,  False, '',           None,     '/**',    False, None,       True)),
        ('change_profile -> "ba r",',                   exp(False, False, False, '',           None,     None,     True,  'ba r',     False)),
        ('audit allow change_profile -> "ba r",',       exp(True,  True,  False, '',           None,     None,     True,  'ba r',     False)),
    )

    def _run_test(self, rawrule, expected):
        self.assertTrue(ChangeProfileRule.match(rawrule))
        obj = ChangeProfileRule.create_instance(rawrule)
        self.assertEqual(rawrule.strip(), obj.raw_rule)
        self._compare_obj(obj, expected)


class ChangeProfileTestParseInvalid(ChangeProfileTest):
    tests = (
        ('change_profile -> ,',            AppArmorException),
        ('change_profile foo -> ,',        AppArmorException),
        ('change_profile notsafe,',        AppArmorException),
        ('change_profile safety -> /bar,', AppArmorException),
    )

    def _run_test(self, rawrule, expected):
        self.assertFalse(ChangeProfileRule.match(rawrule))
        with self.assertRaises(expected):
            ChangeProfileRule.create_instance(rawrule)


class ChangeProfileTestParseFromLog(ChangeProfileTest):
    def test_change_profile_from_log(self):
        parser = ReadLog('', '', '')

        event = 'type=AVC msg=audit(1428699242.551:386): apparmor="DENIED" operation="change_profile" profile="/foo/changeprofile" pid=3459 comm="changeprofile" target="/foo/rename"'

        # libapparmor doesn't understand this log format (from JJ)
        # event = '[   97.492562] audit: type=1400 audit(1431116353.523:77): apparmor="DENIED" operation="change_profile" profile="/foo/changeprofile" pid=3459 comm="changeprofile" target="/foo/rename"'

        parsed_event = parser.parse_event(event)

        self.assertEqual(parsed_event, {
            'request_mask': None,
            'denied_mask': None,
            'error_code': 0,
            'magic_token': 0,
            'parent': 0,
            'profile': '/foo/changeprofile',
            'operation': 'change_profile',
            'resource': None,
            'info': None,
            'aamode': 'REJECTING',
            'time': 1428699242,
            'active_hat': None,
            'pid': 3459,
            'task': 0,
            'attr': None,
            'name2': '/foo/rename',  # target
            'name': None,
            'family': None,
            'protocol': None,
            'sock_type': None,
            'class': None,
        })

        obj = ChangeProfileRule(None, ChangeProfileRule.ALL, parsed_event['name2'], log_event=parsed_event)

        #             audit  allow  deny  comment  execmode  execcond  all?   targetprof    all?
        expected = exp(False, False, False, '',    None,     None,    True, '/foo/rename', False)

        self._compare_obj(obj, expected)

        self.assertEqual(obj.get_raw(1), '  change_profile -> /foo/rename,')


class ChangeProfileFromInit(ChangeProfileTest):
    tests = (
        # ChangeProfileRule object                                                      audit  allow  deny   comment  execmode  execcond  all?   targetprof  all?
        (ChangeProfileRule(None,     '/foo',                '/bar', deny=True),     exp(False, False, True,  '',      None,     '/foo',   False, '/bar',     False)),
        (ChangeProfileRule(None,     '/foo',                '/bar'),                exp(False, False, False, '',      None,     '/foo',   False, '/bar',     False)),
        (ChangeProfileRule('safe',   '/foo',                '/bar'),                exp(False, False, False, '',      'safe',   '/foo',   False, '/bar',     False)),
        (ChangeProfileRule('unsafe', '/foo',                '/bar'),                exp(False, False, False, '',      'unsafe', '/foo',   False, '/bar',     False)),
        (ChangeProfileRule(None,     '/foo',                ChangeProfileRule.ALL), exp(False, False, False, '',      None,     '/foo',   False, None,       True)),
        (ChangeProfileRule(None,     ChangeProfileRule.ALL, '/bar'),                exp(False, False, False, '',      None,     None,     True,  '/bar',     False)),
        (ChangeProfileRule(None,     ChangeProfileRule.ALL, ChangeProfileRule.ALL), exp(False, False, False, '',      None,     None,     True,  None,       True)),
    )

    def _run_test(self, obj, expected):
        self._compare_obj(obj, expected)


class InvalidChangeProfileInit(AATest):
    tests = (
        # init params               expected exception
        ((None,    '/foo', ''),     AppArmorBug),  # empty targetprofile
        ((None,    '',     '/bar'), AppArmorBug),  # empty execcond
        ((None,    '    ', '/bar'), AppArmorBug),  # whitespace execcond
        ((None,    '/foo', '   '),  AppArmorBug),  # whitespace targetprofile
        ((None,    'xyxy', '/bar'), AppArmorException),  # invalid execcond
        ((None,    dict(), '/bar'), AppArmorBug),  # wrong type for execcond
        ((None,    None,   '/bar'), AppArmorBug),  # wrong type for execcond
        ((None,    '/foo', dict()), AppArmorBug),  # wrong type for targetprofile
        ((None,    '/foo', None),   AppArmorBug),  # wrong type for targetprofile
        (('maybe', '/foo', '/bar'), AppArmorBug),  # invalid keyword for execmode
    )

    def _run_test(self, params, expected):
        with self.assertRaises(expected):
            ChangeProfileRule(*params)

    def test_missing_params_1(self):
        with self.assertRaises(TypeError):
            ChangeProfileRule()

    def test_missing_params_2(self):
        with self.assertRaises(TypeError):
            ChangeProfileRule('inet')


class InvalidChangeProfileTest(AATest):
    def _check_invalid_rawrule(self, rawrule):
        obj = None
        self.assertFalse(ChangeProfileRule.match(rawrule))
        with self.assertRaises(AppArmorException):
            obj = ChangeProfileRule.create_instance(rawrule)

        self.assertIsNone(obj, 'ChangeProfileRule handed back an object unexpectedly')

    def test_invalid_net_missing_comma(self):
        self._check_invalid_rawrule('change_profile')  # missing comma

    def test_invalid_net_non_ChangeProfileRule(self):
        self._check_invalid_rawrule('dbus,')  # not a change_profile rule

    def test_empty_net_data_1(self):
        obj = ChangeProfileRule(None, '/foo', '/bar')
        obj.execcond = ''
        # no execcond set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)

    def test_empty_net_data_2(self):
        obj = ChangeProfileRule(None, '/foo', '/bar')
        obj.targetprofile = ''
        # no targetprofile set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)


class WriteChangeProfileTestAATest(AATest):
    tests = (
        #  raw rule                                                         clean rule
        ('     change_profile         ,    # foo        ',                  'change_profile, # foo'),
        ('    audit     change_profile /foo,',                              'audit change_profile /foo,'),
        ('   deny change_profile         /foo      -> bar,# foo bar',       'deny change_profile /foo -> bar, # foo bar'),
        ('   deny change_profile         /foo      ,# foo bar',             'deny change_profile /foo, # foo bar'),
        ('   allow change_profile   ->    /bar     ,# foo bar',             'allow change_profile -> /bar, # foo bar'),
        ('   allow change_profile   unsafe  /** ->    /bar     ,# foo bar', 'allow change_profile unsafe /** -> /bar, # foo bar'),
        ('   allow change_profile   "/fo o" ->    "/b ar",',                'allow change_profile "/fo o" -> "/b ar",'),
        (' priority=9     audit deny     change_profile /foo -> baz,',      'priority=9 audit deny change_profile /foo -> baz,'),
        (' priority = 0   audit deny     change_profile /foo -> baz,',      'priority=0 audit deny change_profile /foo -> baz,'),
        (' priority=-54   audit deny     change_profile /foo -> baz,',      'priority=-54 audit deny change_profile /foo -> baz,'),
        (' priority=+42   audit deny     change_profile /foo -> baz,',      'priority=42 audit deny change_profile /foo -> baz,'),
    )

    def _run_test(self, rawrule, expected):
        self.assertTrue(ChangeProfileRule.match(rawrule))
        obj = ChangeProfileRule.create_instance(rawrule)
        clean = obj.get_clean()
        raw = obj.get_raw()

        self.assertEqual(expected.strip(), clean, 'unexpected clean rule')
        self.assertEqual(rawrule.strip(), raw, 'unexpected raw rule')

    def test_write_manually(self):
        obj = ChangeProfileRule(None, '/foo', 'bar', allow_keyword=True)

        expected = '    allow change_profile /foo -> bar,'

        self.assertEqual(expected, obj.get_clean(2), 'unexpected clean rule')
        self.assertEqual(expected, obj.get_raw(2), 'unexpected raw rule')


class ChangeProfileCoveredTest(AATest):
    def _run_test(self, param, expected):
        obj = ChangeProfileRule.create_instance(self.rule)
        check_obj = ChangeProfileRule.create_instance(param)

        self.assertTrue(ChangeProfileRule.match(param))

        self.assertEqual(obj.is_equal(check_obj), expected[0], 'Mismatch in is_equal, expected {}'.format(expected[0]))
        self.assertEqual(obj.is_equal(check_obj, True), expected[1], 'Mismatch in is_equal/strict, expected {}'.format(expected[1]))

        self.assertEqual(obj.is_covered(check_obj), expected[2], 'Mismatch in is_covered, expected {}'.format(expected[2]))
        self.assertEqual(obj.is_covered(check_obj, True, True), expected[3], 'Mismatch in is_covered/exact, expected {}'.format(expected[3]))


class ChangeProfileCoveredTest_01(ChangeProfileCoveredTest):
    rule = 'change_profile /foo,'

    tests = (
        #   rule                                       equal  strict equal    covered     covered exact
        ('           change_profile,',                (False, False,          False,      False)),
        ('           change_profile /foo,',           (True,  True,           True,       True)),
        ('           change_profile safe /foo,',      (True,  False,          True,       True)),
        ('           change_profile unsafe /foo,',    (False, False,          False,      False)),
        ('           change_profile /foo, # comment', (True,  False,          True,       True)),
        ('     allow change_profile /foo,',           (True,  False,          True,       True)),
        ('           change_profile     /foo,',       (True,  False,          True,       True)),
        ('           change_profile /foo -> /bar,',   (False, False,          True,       True)),
        ('           change_profile /foo -> bar,',    (False, False,          True,       True)),
        ('audit      change_profile /foo,',           (False, False,          False,      False)),
        ('audit      change_profile,',                (False, False,          False,      False)),
        ('           change_profile /asdf,',          (False, False,          False,      False)),
        ('           change_profile -> /bar,',        (False, False,          False,      False)),
        ('audit deny change_profile /foo,',           (False, False,          False,      False)),
        ('      deny change_profile /foo,',           (False, False,          False,      False)),
    )


class ChangeProfileCoveredTest_02(ChangeProfileCoveredTest):
    rule = 'audit change_profile /foo,'

    tests = (
        #   rule                                     equal  strict equal covered  covered exact
        ('      change_profile /foo,',              (False, False,       True,    False)),
        ('audit change_profile /foo,',              (True,  True,        True,    True)),
        ('      change_profile /foo -> /bar,',      (False, False,       True,    False)),
        ('      change_profile safe /foo -> /bar,', (False, False,       True,    False)),
        ('audit change_profile /foo -> /bar,',      (False, False,       True,    True)),  # XXX is "covered exact" correct here?
        ('      change_profile,',                   (False, False,       False,   False)),
        ('audit change_profile,',                   (False, False,       False,   False)),
        ('      change_profile -> /bar,',           (False, False,       False,   False)),
    )


class ChangeProfileCoveredTest_03(ChangeProfileCoveredTest):
    rule = 'change_profile /foo -> /bar,'

    tests = (
        #   rule                                equal  strict equal  covered  covered exact
        ('      change_profile /foo -> /bar,', (True,  True,         True,    True)),
        ('allow change_profile /foo -> /bar,', (True,  False,        True,    True)),
        ('      change_profile /foo,',         (False, False,        False,   False)),
        ('      change_profile,',              (False, False,        False,   False)),
        ('      change_profile /foo -> /xyz,', (False, False,        False,   False)),
        ('audit change_profile,',              (False, False,        False,   False)),
        ('audit change_profile /foo -> /bar,', (False, False,        False,   False)),
        ('      change_profile      -> /bar,', (False, False,        False,   False)),
        ('      change_profile,',              (False, False,        False,   False)),
    )


class ChangeProfileCoveredTest_04(ChangeProfileCoveredTest):
    rule = 'change_profile,'

    tests = (
        #   rule                                equal  strict equal  covered  covered exact
        ('      change_profile,',              (True,  True,         True,    True)),
        ('allow change_profile,',              (True,  False,        True,    True)),
        ('      change_profile /foo,',         (False, False,        True,    True)),
        ('      change_profile /xyz -> bar,',  (False, False,        True,    True)),
        ('      change_profile -> /bar,',      (False, False,        True,    True)),
        ('      change_profile /foo -> /bar,', (False, False,        True,    True)),
        ('audit change_profile,',              (False, False,        False,   False)),
        ('deny  change_profile,',              (False, False,        False,   False)),
    )


class ChangeProfileCoveredTest_05(ChangeProfileCoveredTest):
    rule = 'deny change_profile /foo,'

    tests = (
        #   rule                             equal  strict equal  covered  covered exact
        ('      deny change_profile /foo,', (True,  True,         True,    True)),
        ('audit deny change_profile /foo,', (False, False,        False,   False)),
        ('           change_profile /foo,', (False, False,        False,   False)),  # XXX should covered be true here?
        ('      deny change_profile /bar,', (False, False,        False,   False)),
        ('      deny change_profile,',      (False, False,        False,   False)),
    )


class ChangeProfileCoveredTest_06(ChangeProfileCoveredTest):
    rule = 'change_profile safe /foo,'

    tests = (
        #   rule                             equal  strict equal  covered  covered exact
        ('      deny change_profile /foo,', (False, False,        False,   False)),
        ('audit deny change_profile /foo,', (False, False,        False,   False)),
        ('           change_profile /foo,', (True,  False,        True,    True)),
        ('      deny change_profile /bar,', (False, False,        False,   False)),
        ('      deny change_profile,',      (False, False,        False,   False)),
    )


class ChangeProfileCoveredTest_Invalid(AATest):
    def test_borked_obj_is_covered_1(self):
        obj = ChangeProfileRule.create_instance('change_profile /foo,')

        testobj = ChangeProfileRule(None, '/foo', '/bar')
        testobj.execcond = ''

        with self.assertRaises(AppArmorBug):
            obj.is_covered(testobj)

    def test_borked_obj_is_covered_2(self):
        obj = ChangeProfileRule.create_instance('change_profile /foo,')

        testobj = ChangeProfileRule(None, '/foo', '/bar')
        testobj.targetprofile = ''

        with self.assertRaises(AppArmorBug):
            obj.is_covered(testobj)

    def test_invalid_is_covered(self):
        raw_rule = 'change_profile /foo,'

        class SomeOtherClass(ChangeProfileRule):
            pass

        obj = ChangeProfileRule.create_instance(raw_rule)
        testobj = SomeOtherClass.create_instance(raw_rule)  # different type
        with self.assertRaises(AppArmorBug):
            obj.is_covered(testobj)

    def test_invalid_is_equal(self):
        raw_rule = 'change_profile -> /bar,'

        class SomeOtherClass(ChangeProfileRule):
            pass

        obj = ChangeProfileRule.create_instance(raw_rule)
        testobj = SomeOtherClass.create_instance(raw_rule)  # different type
        with self.assertRaises(AppArmorBug):
            obj.is_equal(testobj)


class ChangeProfileLogprofHeaderTest(AATest):
    tests = (
        ('change_profile,',                       [                                                        _('Exec Condition'), _('ALL'),  _('Target Profile'), _('ALL')]),  # noqa: E201
        ('change_profile -> /bin/ping,',          [                                                        _('Exec Condition'), _('ALL'),  _('Target Profile'), '/bin/ping']),  # noqa: E201
        ('change_profile /bar -> /bin/bar,',      [                                                        _('Exec Condition'), '/bar',    _('Target Profile'), '/bin/bar']),  # noqa: E201
        ('change_profile safe /foo,',             [                              _('Exec Mode'), 'safe',   _('Exec Condition'), '/foo',    _('Target Profile'), _('ALL')]),  # noqa: E201
        ('audit change_profile -> /bin/ping,',    [_('Qualifier'), 'audit',                                _('Exec Condition'), _('ALL'),  _('Target Profile'), '/bin/ping']),
        ('deny change_profile /bar -> /bin/bar,', [_('Qualifier'), 'deny',                                 _('Exec Condition'), '/bar',    _('Target Profile'), '/bin/bar']),
        ('allow change_profile unsafe /foo,',     [_('Qualifier'), 'allow',      _('Exec Mode'), 'unsafe', _('Exec Condition'), '/foo',    _('Target Profile'), _('ALL')]),
        ('audit deny change_profile,',            [_('Qualifier'), 'audit deny',                           _('Exec Condition'), _('ALL'),  _('Target Profile'), _('ALL')]),
    )

    def _run_test(self, params, expected):
        obj = ChangeProfileRule.create_instance(params)
        self.assertEqual(obj.logprof_header(), expected)


# --- tests for ChangeProfileRuleset --- #

class ChangeProfileRulesTest(AATest):
    def test_empty_ruleset(self):
        ruleset = ChangeProfileRuleset()
        ruleset_2 = ChangeProfileRuleset()
        self.assertEqual([], ruleset.get_raw(2))
        self.assertEqual([], ruleset.get_clean(2))
        self.assertEqual([], ruleset_2.get_raw(2))
        self.assertEqual([], ruleset_2.get_clean(2))

    def test_ruleset_1(self):
        ruleset = ChangeProfileRuleset()
        rules = (
            'change_profile -> /bar,',
            'change_profile /foo,',
        )

        expected_raw = [
            'change_profile -> /bar,',
            'change_profile /foo,',
            '',
        ]

        expected_clean = [
            'change_profile -> /bar,',
            'change_profile /foo,',
            '',
        ]

        for rule in rules:
            ruleset.add(ChangeProfileRule.create_instance(rule))

        self.assertEqual(expected_raw, ruleset.get_raw())
        self.assertEqual(expected_clean, ruleset.get_clean())

    def test_ruleset_2(self):
        ruleset = ChangeProfileRuleset()
        rules = (
            'change_profile /foo -> /bar,',
            'allow change_profile /asdf,',
            'deny change_profile -> xy, # example comment',
        )

        expected_raw = [
            '  change_profile /foo -> /bar,',
            '  allow change_profile /asdf,',
            '  deny change_profile -> xy, # example comment',
            '',
        ]

        expected_clean = [
            '  deny change_profile -> xy, # example comment',
            '',
            '  allow change_profile /asdf,',
            '  change_profile /foo -> /bar,',
            '',
        ]

        for rule in rules:
            ruleset.add(ChangeProfileRule.create_instance(rule))

        self.assertEqual(expected_raw, ruleset.get_raw(1))
        self.assertEqual(expected_clean, ruleset.get_clean(1))


class ChangeProfileGlobTestAATest(AATest):
    def setUp(self):
        self.ruleset = ChangeProfileRuleset()

    def test_glob_1(self):
        self.assertEqual(self.ruleset.get_glob('change_profile /foo,'), 'change_profile,')

    # not supported or used yet, glob behaviour not decided yet
    # def test_glob_2(self):
    #     self.assertEqual(self.ruleset.get_glob('change_profile /foo -> /bar,'), 'change_profile -> /bar,')

    def test_glob_ext(self):
        with self.assertRaises(NotImplementedError):
            # get_glob_ext is not available for change_profile rules
            self.ruleset.get_glob_ext('change_profile /foo -> /bar,')


class ChangeProfileDeleteTestAATest(AATest):
    pass


setup_all_loops(__name__)
if __name__ == '__main__':
    unittest.main(verbosity=1)



#!/usr/bin/python3
#
#    Copyright (C) 2012 Canonical Ltd.
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License published by the Free Software Foundation.
#
#    Written by Steve Beattie <steve@nxnw.org>, based on work by
#    Christian Boltz <apparmor@cboltz.de>

import re
import subprocess
import sys

# dangerous capabilities
danger_caps = ("audit_control",
               "audit_write",
               "mac_override",
               "mac_admin",
               "setfcap",
               "sys_admin",
               "sys_module",
               "sys_rawio")


def cmd(command, input=None, stderr=subprocess.STDOUT, stdout=subprocess.PIPE, stdin=None, timeout=None):
    """Try to execute given command (array) and return its stdout, or
    return a textual error if it failed."""

    try:
        sp = subprocess.Popen(command, stdin=stdin, stdout=stdout, stderr=stderr, close_fds=True, universal_newlines=True)
    except OSError as ex:
        return 127, str(ex)

    out, outerr = sp.communicate(input)

    # Handle redirection of stdout
    if out is None:
        out = ''
    # Handle redirection of stderr
    if outerr is None:
        outerr = ''
    return sp.returncode, out, outerr


# get capabilities list
(rc, output, outerr) = cmd(('../../common/list_capabilities.sh',))
if rc != 0:
    sys.stderr.write("make list_capabilities failed: " + output + outerr)
    sys.exit(rc)

capabilities = re.sub('CAP_', '', output.strip()).lower().split('\n')
benign_caps = []
for cap in capabilities:
    if cap not in danger_caps:
        benign_caps.append(cap)

# get network protos list
(rc, output, outerr) = cmd(('../../common/list_af_names.sh',))
if rc != 0:
    sys.stderr.write("make list_af_names failed: " + output + outerr)
    sys.exit(rc)

af_names = []
af_pairs = re.sub('AF_', '', output.strip()).lower().split(",")
for af_pair in af_pairs:
    af_name = af_pair.lstrip().split(" ")[0]
    # skip max af name definition
    if af_name and af_name != "max":
        af_names.append(af_name)

# TODO: does a "debug" flag exist? Listed in apparmor.vim.in sdFlagKey,
# but not in aa_flags...
# -> currently (2011-01-11) not, but might come back

aa_network_types = r'\s+tcp|\s+udp|\s+icmp'

aa_flags = ('complain',
            'unconfined',
            'audit',
            'attach_disconnected',
            'no_attach_disconnected',
            'chroot_attach',
            'chroot_no_attach',
            'chroot_relative',
            'namespace_relative',
            'mediate_deleted',
            'delegate_deleted')

filename = r'(\/|\@\{\S*\})\S*'

aa_regex_map = {
    'FILENAME':         filename,
    'FILE':             r'\v^\s*(audit\s+)?(deny\s+|allow\s+)?(owner\s+|other\s+)?' + filename + r'\s+',  # Start of a file rule
                        # (whitespace_+_, owner etc. flag_?_, filename pattern, whitespace_+_)
    'DENYFILE':         r'\v^\s*(audit\s+)?deny\s+(owner\s+|other\s+)?' + filename + r'\s+',  # deny, otherwise like FILE
    'auditdenyowner':   r'(audit\s+)?(deny\s+|allow\s+)?(owner\s+|other\s+)?',
    'audit_DENY_owner': r'(audit\s+)?deny\s+(owner\s+|other\s+)?',  # must include "deny", otherwise like auditdenyowner
    'auditdeny':        r'(audit\s+)?(deny\s+|allow\s+)?',
    'EOL':              r'\s*,(\s*$|(\s*#.*$)\@=)',  # End of a line (whitespace_?_, comma, whitespace_?_ comment.*)
    'TRANSITION':       r'(\s+-\>\s+\S+)?',
    'sdKapKey':         " ".join(benign_caps),
    'sdKapKeyDanger':   " ".join(danger_caps),
    'sdKapKeyRegex':    "|".join(capabilities),
    'sdNetworkType':    aa_network_types,
    'sdNetworkProto':   "|".join(af_names),
    'flags':            r'((flags\s*\=\s*)?\(\s*(' + '|'.join(aa_flags) + r')(\s*,\s*(' + '|'.join(aa_flags) + r'))*\s*\)\s+)',
}


def my_repl(matchobj):
    matchobj.group(1)
    if matchobj.group(1) in aa_regex_map:
        return aa_regex_map[matchobj.group(1)]

    return matchobj.group(0)


def create_file_rule(highlighting, permissions, comment, denyrule=0):

    if denyrule == 0:
        keywords = '@@auditdenyowner@@'
    else:
        keywords = '@@audit_DENY_owner@@'  # TODO: not defined yet, will be '(audit\s+)?deny\s+(owner\s+)?'

    sniplet = ''
    sniplet = sniplet + "\n" + '" ' + comment + "\n"

    prefix = r'syn match  ' + highlighting + r' /\v^\s*' + keywords
    suffix = r'@@EOL@@/ contains=sdGlob,sdComment nextgroup=@sdEntry,sdComment,sdError,sdInclude' + "\n"
    # filename without quotes
    sniplet = sniplet + prefix + r'@@FILENAME@@\s+' + permissions + suffix
    # filename with quotes
    sniplet = sniplet + prefix + r'"@@FILENAME@@"\s+' + permissions + suffix
    # filename without quotes, reverse syntax
    sniplet = sniplet + prefix + permissions + r'\s+@@FILENAME@@' + suffix
    # filename with quotes, reverse syntax
    sniplet = sniplet + prefix + permissions + r'\s+"@@FILENAME@@"+' + suffix

    return sniplet


filerule = ''
filerule = filerule + create_file_rule('sdEntryWriteExec ', r'(l|r|w|a|m|k|[iuUpPcC]x)+@@TRANSITION@@', 'write + exec/mmap - danger! (known bug: accepts aw to keep things simple)')
filerule = filerule + create_file_rule('sdEntryUX',  r'(r|m|k|ux|pux)+@@TRANSITION@@',  'ux(mr) - unconstrained entry, flag the line red. also includes pux which is unconstrained if no profile exists')
filerule = filerule + create_file_rule('sdEntryUXe', r'(r|m|k|Ux|PUx)+@@TRANSITION@@',  'Ux(mr) and PUx(mr) - like ux + clean environment')
filerule = filerule + create_file_rule('sdEntryPX',  r'(r|m|k|px|cx|pix|cix)+@@TRANSITION@@',  'px/cx/pix/cix(mrk) - standard exec entry, flag the line blue')
filerule = filerule + create_file_rule('sdEntryPXe', r'(r|m|k|Px|Cx|Pix|Cix)+@@TRANSITION@@', 'Px/Cx/Pix/Cix(mrk) - like px/cx + clean environment')
filerule = filerule + create_file_rule('sdEntryIX',  r'(r|m|k|ix)+',  'ix(mr) - standard exec entry, flag the line green')
filerule = filerule + create_file_rule('sdEntryM',   r'(r|m|k)+',  'mr - mmap with PROT_EXEC')

filerule = filerule + create_file_rule('sdEntryM',   r'(r|m|k|x)+',  'special case: deny x is allowed (does not need to be ix, px, ux or cx)', 1)
# syn match  sdEntryM /@@DENYFILE@@(r|m|k|x)+@@EOL@@/ contains=sdGlob,sdComment nextgroup=@sdEntry,sdComment,sdError,sdInclude


filerule = filerule + create_file_rule('sdError',    r'\S*(w\S*a|a\S*w)\S*',  'write + append is an error')
filerule = filerule + create_file_rule('sdEntryW',   r'(l|r|w|k)+',  'write entry, flag the line yellow')
filerule = filerule + create_file_rule('sdEntryW',   r'(l|r|a|k)+',  'append entry, flag the line yellow')
filerule = filerule + create_file_rule('sdEntryK',   r'[rlk]+',  'read entry + locking, currently no highlighting')
filerule = filerule + create_file_rule('sdEntryR',   r'[rl]+',  'read entry, no highlighting')

# " special case: deny x is allowed (doesn't need to be ix, px, ux or cx)
# syn match  sdEntryM /@@DENYFILE@@(r|m|k|x)+@@EOL@@/ contains=sdGlob,sdComment nextgroup=@sdEntry,sdComment,sdError,sdInclude

# " TODO: Support filenames enclosed in quotes ("/home/foo/My Documents/") - ideally by only allowing quotes pair-wise


regex = "@@(" + "|".join(aa_regex_map) + ")@@"

sys.stdout.write('" generated from apparmor.vim.in by create-apparmor.vim.py\n')
sys.stdout.write('" do not edit this file - edit apparmor.vim.in or create-apparmor.vim.py instead' + "\n\n")

with open("apparmor.vim.in") as template:
    for line in template:
        line = re.sub(regex, my_repl, line.rstrip())
        sys.stdout.write(line + '\n')

sys.stdout.write("\n\n\n\n")

sys.stdout.write('" file rules added with create_file_rule()\n')
sys.stdout.write(re.sub(regex, my_repl, filerule) + '\n')



#! /usr/bin/python3
# ------------------------------------------------------------------
#
#    Copyright (C) 2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License published by the Free Software Foundation.
#
# ------------------------------------------------------------------

import os
import unittest

import apparmor.aa as apparmor
from apparmor.common import AppArmorException, is_skippable_file, open_file_read
from common_test import AATest, setup_aa, setup_all_loops

# This testcase will parse all parser/tst/simple_tests with parse_profile_data(),
# except the files listed in one of the arrays below.
#
# Files listed in skip_startswith will be completely skipped.
# Files listed in the other arrays will be checked, but with the opposite of the expected result.


# XXX tests listed here will be *** SKIPPED *** XXX
skip_startswith = (
    # the tools don't check for conflicting x permissions (yet?)
    'generated_x/conflict-',
    'generated_x/ambiguous-',
    'generated_x/dominate-',

    # 'safe' and 'unsafe' keywords
    'generated_perms_safe/',

    # Pux and Cux (which actually mean PUx and CUx) get rejected by the tools
    'generated_x/exact-',
)

# testcases that should raise an exception, but don't
exception_not_raised = (
    # most abi/bad_* aren't detected as bad by the basic implementation in the tools
    'abi/bad_10.sd',
    'abi/bad_11.sd',
    'abi/bad_12.sd',

    # interesting[tm] profile name
    'change_hat/bad_parsing.sd',

    'dbus/bad_regex_04.sd',
    'dbus/bad_regex_05.sd',
    'dbus/bad_regex_06.sd',
    'file/bad_re_brace_1.sd',
    'file/bad_re_brace_2.sd',
    'file/bad_re_brace_3.sd',

    # The tools don't detect conflicting change_profile exec modes
    'change_profile/onx_conflict_unsafe1.sd',
    'change_profile/onx_conflict_unsafe2.sd',

    # duplicated conditionals aren't detected by the tools
    'generated_dbus/duplicated-conditionals-45127.sd',
    'generated_dbus/duplicated-conditionals-45131.sd',
    'generated_dbus/duplicated-conditionals-45124.sd',
    'generated_dbus/duplicated-conditionals-45130.sd',
    'generated_dbus/duplicated-conditionals-45125.sd',
    'generated_dbus/duplicated-conditionals-45128.sd',
    'generated_dbus/duplicated-conditionals-45129.sd',

    'dbus/bad_modifier_2.sd',
    'dbus/bad_regex_01.sd',
    'dbus/bad_regex_02.sd',
    'dbus/bad_regex_03.sd',
    'dbus/bad_regex_04.sd',
    'dbus/bad_regex_05.sd',
    'dbus/bad_regex_06.sd',
    'file/bad_re_brace_1.sd',
    'file/bad_re_brace_2.sd',
    'file/bad_re_brace_3.sd',

    # We do not check that options are compatible
    'mount/bad_opt_29.sd',
    'mount/bad_opt_30.sd',
    'mount/bad_opt_31.sd',
    'mount/bad_1.sd',
    'mount/bad_2.sd',
    'mount/bad_3.sd',
    'mount/bad_4.sd',

    'profile/flags/flags_bad10.sd',
    'profile/flags/flags_bad11.sd',
    'profile/flags/flags_bad12.sd',
    'profile/flags/flags_bad13.sd',
    'profile/flags/flags_bad15.sd',
    'profile/flags/flags_bad18.sd',
    'profile/flags/flags_bad19.sd',
    'profile/flags/flags_bad2.sd',
    'profile/flags/flags_bad3.sd',
    'profile/flags/flags_bad4.sd',
    'profile/flags/flags_bad5.sd',
    'profile/flags/flags_bad6.sd',
    'profile/flags/flags_bad7.sd',
    'profile/flags/flags_bad8.sd',
    'profile/flags/flags_bad_debug_1.sd',
    'profile/flags/flags_bad_debug_2.sd',
    'profile/flags/flags_bad_debug_3.sd',
    # detection of conflicting flags not supported
    'profile/flags/flags_bad30.sd',
    'profile/flags/flags_bad31.sd',
    'profile/flags/flags_bad32.sd',
    'profile/flags/flags_bad33.sd',
    'profile/flags/flags_bad34.sd',
    'profile/flags/flags_bad35.sd',
    'profile/flags/flags_bad36.sd',
    'profile/flags/flags_bad37.sd',
    'profile/flags/flags_bad38.sd',
    'profile/flags/flags_bad39.sd',
    'profile/flags/flags_bad40.sd',
    'profile/flags/flags_bad41.sd',
    'profile/flags/flags_bad42.sd',
    'profile/flags/flags_bad43.sd',
    'profile/flags/flags_bad44.sd',
    'profile/flags/flags_bad45.sd',
    'profile/flags/flags_bad46.sd',
    'profile/flags/flags_bad47.sd',
    'profile/flags/flags_bad48.sd',
    'profile/flags/flags_bad49.sd',
    'profile/flags/flags_bad50.sd',
    'profile/flags/flags_bad51.sd',
    'profile/flags/flags_bad52.sd',
    'profile/flags/flags_bad53.sd',
    'profile/flags/flags_bad54.sd',
    'profile/flags/flags_bad55.sd',
    'profile/flags/flags_bad56.sd',
    'profile/flags/flags_bad64.sd',
    'profile/flags/flags_bad65.sd',
    'profile/flags/flags_bad66.sd',
    'profile/flags/flags_bad67.sd',
    'profile/flags/flags_bad68.sd',
    'profile/flags/flags_bad69.sd',
    'profile/flags/flags_bad70.sd',
    'profile/flags/flags_bad71.sd',
    'profile/flags/flags_bad72.sd',
    'profile/flags/flags_bad73.sd',
    'profile/flags/flags_bad74.sd',
    'profile/flags/flags_bad75.sd',
    'profile/flags/flags_bad76.sd',
    'profile/flags/flags_bad77.sd',
    'profile/flags/flags_bad78.sd',
    'profile/flags/flags_bad79.sd',
    'profile/flags/flags_bad80.sd',
    'profile/flags/flags_bad81.sd',
    'profile/flags/flags_bad82.sd',
    'profile/flags/flags_bad83.sd',
    'profile/flags/flags_bad84.sd',
    'profile/flags/flags_bad85.sd',
    'profile/flags/flags_bad86.sd',
    # flags=(error=EXX)
    'profile/flags/flags_bad87.sd',
    'profile/flags/flags_bad88.sd',
    'profile/flags/flags_bad89.sd',

    'profile/flags/flags_bad_disconnected_path1.sd',
    'profile/flags/flags_bad_disconnected_path2.sd',
    'profile/flags/flags_bad_disconnected_path3.sd',
    'profile/flags/flags_bad_disconnected_path4.sd',
    'profile/flags/flags_bad_disconnected_path5.sd',
    'profile/profile_ns_bad8.sd',  # 'profile :ns/t' without terminating ':'
    'ptrace/bad_10.sd',  # peer with invalid regex
    'signal/bad_21.sd',  # invalid regex

    # Invalid regexes
    'unix/bad_regex_01.sd',
    'unix/bad_regex_02.sd',
    'unix/bad_regex_03.sd',
    'unix/bad_regex_04.sd',

    'unix/bad_modifier_2.sd',  # We do not check for duplicated keywords
    'unix/bad_bind_2.sd',  # We do not check bind coherency

    'vars/vars_bad_3.sd',
    'vars/vars_bad_4.sd',
    'vars/vars_bad_5.sd',
    'vars/vars_dbus_bad_01.sd',
    'vars/vars_dbus_bad_02.sd',
    'vars/vars_dbus_bad_03.sd',
    'vars/vars_dbus_bad_04.sd',
    'vars/vars_dbus_bad_05.sd',
    'vars/vars_dbus_bad_06.sd',
    'vars/vars_dbus_bad_07.sd',
    'vars/vars_file_evaluation_8.sd',

    # profile name in var doesn't start with /
    'vars/vars_profile_name_bad_1.sd',

    # profile name is undefined variable
    'vars/vars_profile_name_bad_2.sd',
    'vars/vars_profile_name_23.sd',

    # attachment in var doesn't start with /
    'vars/vars_profile_name_07.sd',
    'vars/vars_profile_name_08.sd',
    'vars/vars_profile_name_12.sd',
    'vars/vars_profile_name_13.sd',
    'vars/vars_profile_name_15.sd',
    'vars/vars_profile_name_22.sd',

    'vars/vars_recursion_1.sd',
    'vars/vars_recursion_2.sd',
    'vars/vars_recursion_3.sd',
    'vars/vars_recursion_4.sd',
    'vars/vars_simple_assignment_3.sd',
    'vars/vars_simple_assignment_8.sd',
    'vars/vars_simple_assignment_9.sd',
    'xtrans/simple_bad_conflicting_x_10.sd',
    'xtrans/simple_bad_conflicting_x_11.sd',
    'xtrans/simple_bad_conflicting_x_12.sd',
    'xtrans/simple_bad_conflicting_x_13.sd',
    'xtrans/simple_bad_conflicting_x_2.sd',
    'xtrans/simple_bad_conflicting_x_4.sd',
    'xtrans/simple_bad_conflicting_x_6.sd',
    'xtrans/simple_bad_conflicting_x_8.sd',
    'xtrans/x-conflict.sd',

    'network/perms/bad_modifier_2.sd',
)

# testcases with lines that don't match any regex and end up as "unknown line"
unknown_line = (
    # 'other' keyword
    'file/allow/ok_other_1.sd',
    'file/allow/ok_other_2.sd',
    'file/ok_other_1.sd',
    'file/ok_other_2.sd',
    'file/ok_other_3.sd',
    'file/priority/ok_other_1.sd',
    'file/priority/ok_other_2.sd',
    'file/priority/ok_other_3.sd',

    # 'unsafe' keyword
    'file/file/front_perms_ok_1.sd',
    'file/front_perms_ok_1.sd',
    'xtrans/simple_ok_cx_1.sd',
    'file/priority/front_perms_ok_1.sd',
    'file/priority/front_perms_ok_2.sd',

    # permissions before path and owner / audit {...} blocks
    'file/file/owner/ok_1.sd',
    'file/owner/ok_1.sd',
    'profile/entry_mods_audit_ok1.sd',

    # namespace
    'profile/profile_ns_named_ok1.sd',  # profile keyword?
    'profile/profile_ns_named_ok2.sd',  # profile keyword?
    'profile/profile_ns_named_ok3.sd',  # profile keyword?
    'profile/profile_ns_ok1.sd',
    'profile/profile_ns_ok2.sd',
    'profile/profile_ns_ok3.sd',  # profile keyword?
    'profile/re_named_ok4.sd',  # profile keyword
    'profile/re_ok4.sd',

    # multiple rules in one line
    'bare_include_tests/ok_14.sd',
    'bare_include_tests/ok_19.sd',
    'bare_include_tests/ok_64.sd',
    'bare_include_tests/ok_69.sd',

    # include with quoted relative path
    'bare_include_tests/ok_11.sd',
    'bare_include_tests/ok_12.sd',
    'bare_include_tests/ok_13.sd',
    'bare_include_tests/ok_15.sd',
    # include with unquoted relative path
    'bare_include_tests/ok_16.sd',
    'bare_include_tests/ok_17.sd',
    'bare_include_tests/ok_18.sd',
    'bare_include_tests/ok_20.sd',
    # include with quoted relative path with spaces
    'bare_include_tests/ok_26.sd',
    'bare_include_tests/ok_27.sd',
    # include with quoted magic path with spaces
    'bare_include_tests/ok_28.sd',
    'bare_include_tests/ok_29.sd',
    # include with magic path with spaces
    'bare_include_tests/ok_30.sd',
    'bare_include_tests/ok_31.sd',
    # include if exists with quoted relative path
    'bare_include_tests/ok_61.sd',
    'bare_include_tests/ok_62.sd',
    'bare_include_tests/ok_63.sd',
    # include if exists with unquoted relative path
    'bare_include_tests/ok_65.sd',
    'bare_include_tests/ok_66.sd',
    'bare_include_tests/ok_67.sd',
    'bare_include_tests/ok_68.sd',
    'bare_include_tests/ok_70.sd',
    # include if exists with quoted relative path with spaces
    'bare_include_tests/ok_76.sd',
    'bare_include_tests/ok_77.sd',
    # include if exists with quoted magic path with spaces
    'bare_include_tests/ok_78.sd',
    'bare_include_tests/ok_79.sd',
    # include if exists with unquoted magic path with spaces
    'bare_include_tests/ok_80.sd',
    'bare_include_tests/ok_81.sd',
    # include if exists with quoted relative path, non-existing include file
    'bare_include_tests/ok_82.sd',
    'bare_include_tests/ok_84.sd',
    'bare_include_tests/ok_85.sd',
    'bare_include_tests/ok_86.sd',

    # Unsupported \\" in unix AARE
    'unix/ok_regex_03.sd',
    'unix/ok_regex_09.sd',
    'unix/ok_regex_13.sd',
    'unix/ok_regex_19.sd',

)

# testcases with various unexpected failures
syntax_failure = (
    # missing profile keywords
    'profile/re_named_ok2.sd',

    # Syntax Errors caused by boolean conditions (parse_profile_data() gets confused by the closing '}')
    'conditional/defined_1.sd',
    'conditional/defined_2.sd',
    'conditional/else_1.sd',
    'conditional/else_2.sd',
    'conditional/else_3.sd',
    'conditional/else_if_1.sd',
    'conditional/else_if_2.sd',
    'conditional/else_if_3.sd',
    'conditional/else_if_5.sd',
    'conditional/ok_1.sd',
    'conditional/ok_2.sd',
    'conditional/ok_3.sd',
    'conditional/ok_4.sd',
    'conditional/ok_5.sd',
    'conditional/ok_6.sd',
    'conditional/ok_7.sd',
    'conditional/ok_8.sd',
    'conditional/ok_9.sd',
    'conditional/stress_1.sd',

    # unexpected uppercase vs. lowercase in *x rules
    'file/ok_5.sd',  # Invalid mode UX
    'file/ok_2.sd',  # Invalid mode RWM
    'file/ok_4.sd',  # Invalid mode iX
    'file/priority/ok_5.sd',  # Invalid mode UX
    'file/priority/ok_2.sd',  # Invalid mode RWM
    'file/priority/ok_4.sd',  # Invalid mode iX
    'xtrans/simple_ok_pix_1.sd',  # Invalid mode pIx
    'xtrans/simple_ok_pux_1.sd',  # Invalid mode rPux

    # unexpected uppercase vs. lowercase in *x rules - generated_perms_leading directory
    'generated_perms_leading/exact-re-Puxtarget.sd',
    'generated_perms_leading/dominate-ownerCuxtarget2.sd',
    'generated_perms_leading/ambiguous-Cux.sd',
    'generated_perms_leading/dominate-ownerPux.sd',
    'generated_perms_leading/exact-re-ownerPux.sd',
    'generated_perms_leading/overlap-ownerCuxtarget.sd',
    'generated_perms_leading/exact-re-ownerCuxtarget.sd',
    'generated_perms_leading/dominate-Puxtarget2.sd',
    'generated_perms_leading/dominate-ownerCuxtarget.sd',
    'generated_perms_leading/dominate-ownerPuxtarget.sd',
    'generated_perms_leading/ambiguous-Pux.sd',
    'generated_perms_leading/ambiguous-Cuxtarget2.sd',
    'generated_perms_leading/exact-Puxtarget2.sd',
    'generated_perms_leading/ambiguous-ownerCux.sd',
    'generated_perms_leading/exact-ownerPux.sd',
    'generated_perms_leading/ambiguous-ownerPuxtarget.sd',
    'generated_perms_leading/exact-re-ownerPuxtarget.sd',
    'generated_perms_leading/exact-re-Cuxtarget.sd',
    'generated_perms_leading/exact-re-Puxtarget2.sd',
    'generated_perms_leading/dominate-Cux.sd',
    'generated_perms_leading/exact-re-ownerCuxtarget2.sd',
    'generated_perms_leading/ambiguous-ownerCuxtarget.sd',
    'generated_perms_leading/exact-re-Cuxtarget2.sd',
    'generated_perms_leading/ambiguous-Puxtarget.sd',
    'generated_perms_leading/overlap-Puxtarget.sd',
    'generated_perms_leading/ambiguous-Puxtarget2.sd',
    'generated_perms_leading/overlap-Puxtarget2.sd',
    'generated_perms_leading/exact-Puxtarget.sd',
    'generated_perms_leading/overlap-ownerPuxtarget.sd',
    'generated_perms_leading/exact-ownerCuxtarget.sd',
    'generated_perms_leading/exact-re-ownerCux.sd',
    'generated_perms_leading/exact-ownerPuxtarget2.sd',
    'generated_perms_leading/exact-ownerCux.sd',
    'generated_perms_leading/overlap-Cuxtarget2.sd',
    'generated_perms_leading/ambiguous-Cuxtarget.sd',
    'generated_perms_leading/ambiguous-ownerPuxtarget2.sd',
    'generated_perms_leading/dominate-ownerCux.sd',
    'generated_perms_leading/exact-Pux.sd',
    'generated_perms_leading/exact-Cuxtarget.sd',
    'generated_perms_leading/overlap-ownerCuxtarget2.sd',
    'generated_perms_leading/overlap-Pux.sd',
    'generated_perms_leading/overlap-ownerPux.sd',
    'generated_perms_leading/ambiguous-ownerCuxtarget2.sd',
    'generated_perms_leading/exact-re-Cux.sd',
    'generated_perms_leading/exact-re-Pux.sd',
    'generated_perms_leading/overlap-Cuxtarget.sd',
    'generated_perms_leading/exact-re-ownerPuxtarget2.sd',
    'generated_perms_leading/exact-Cuxtarget2.sd',
    'generated_perms_leading/exact-Cux.sd',
    'generated_perms_leading/overlap-Cux.sd',
    'generated_perms_leading/overlap-ownerCux.sd',
    'generated_perms_leading/exact-ownerPuxtarget.sd',
    'generated_perms_leading/dominate-Pux.sd',
    'generated_perms_leading/exact-ownerCuxtarget2.sd',
    'generated_perms_leading/dominate-Puxtarget.sd',
    'generated_perms_leading/ambiguous-ownerPux.sd',
    'generated_perms_leading/overlap-ownerPuxtarget2.sd',
    'generated_perms_leading/dominate-Cuxtarget2.sd',
    'generated_perms_leading/dominate-Cuxtarget.sd',
    'generated_perms_leading/dominate-ownerPuxtarget2.sd',

    # escaping with "\"
    'file/ok_embedded_spaces_4.sd',  # \-escaped space
    'file/file/ok_embedded_spaces_4.sd',  # \-escaped space
    'file/ok_quoted_4.sd',  # quoted string including \"
    'file/priority/ok_quoted_4.sd',  # quoted string including \"
    'file/priority/ok_embedded_spaces_4.sd',  # \-escaped space

    # mount rules with multiple 'options' or 'fstype' are not supported by the tools yet, and when writing them, only the last 'options'/'fstype' would survive.
    # Therefore MountRule intentionally raises an exception when parsing such a rule.
    'mount/ok_opt_87.sd',  # multiple options
    'mount/ok_opt_88.sd',  # multiple fstype

    # misc
    'vars/vars_dbus_12.sd',  # AARE starting with {{ are not handled
    'vars/vars_simple_assignment_12.sd',  # Redefining existing variable @{BAR} ('\' not handled)
    'bare_include_tests/ok_2.sd',  # two #include<...> in one line

    # network port range
    'network/network_ok_17.sd',
)


class TestParseParserTests(AATest):
    tests = []  # filled by parse_test_profiles()

    def _run_test(self, params, expected):
        with open_file_read(params['file']) as f_in:
            data = f_in.readlines()

        if params['disabled']:
            # skip disabled testcases
            return

        if params['tools_wrong']:
            # if the tools are marked as being wrong about a profile, expect the opposite result
            # this makes sure we notice any behaviour change, especially not being wrong anymore
            expected = not expected

        # make sure the profile is known in active_profiles.files
        apparmor.active_profiles.init_file(params['file'])

        if expected:
            apparmor.parse_profile_data(data, params['file'], 0, True)
            apparmor.active_profiles.get_all_merged_variables(params['file'], apparmor.include_list_recursive(apparmor.active_profiles.files[params['file']]))

        else:
            with self.assertRaises(AppArmorException):
                apparmor.parse_profile_data(data, params['file'], 0, True)
                apparmor.active_profiles.get_all_merged_variables(params['file'], apparmor.include_list_recursive(apparmor.active_profiles.files[params['file']]))


def parse_test_profiles(file_with_path):
    """parse the test-related headers of a profile (for example EXRESULT) and add the profile to the set of tests"""
    exresult = None
    exresult_found = False
    description = None
    todo = False
    disabled = False

    with open_file_read(file_with_path) as f_in:
        data = f_in.readlines()

    relfile = os.path.relpath(file_with_path, apparmor.profile_dir)

    for line in data:
        if line.startswith('#=EXRESULT '):
            exresult = line.split()[1]
            if exresult == 'PASS':
                exresult = True
                exresult_found = True
            elif exresult == 'FAIL':
                exresult = False
                exresult_found = True
            else:
                raise Exception('{} contains unknown EXRESULT {}'.format(file_with_path, exresult))

        elif line.upper().startswith('#=DESCRIPTION '):
            description = line.split()[1]

        elif line.rstrip() == '#=TODO':
            todo = True

        elif line.rstrip() == '#=DISABLED':
            disabled = True

    if not exresult_found:
        raise Exception(file_with_path + ' does not contain EXRESULT')

    if not description:
        raise Exception(file_with_path + ' does not contain description')

    tools_wrong = False
    if relfile in exception_not_raised:
        if exresult:
            raise Exception(file_with_path + " listed in exception_not_raised, but has EXRESULT PASS")
        tools_wrong = 'EXCEPTION_NOT_RAISED'
    elif relfile.startswith(skip_startswith):
        return 1  # XXX *** SKIP *** those tests
    elif relfile in unknown_line:
        if not exresult:
            raise Exception(file_with_path + " listed in unknown_line, but has EXRESULT FAIL")
        tools_wrong = 'UNKNOWN_LINE'
    elif relfile in syntax_failure:
        if not exresult:
            raise Exception(file_with_path + " listed in syntax_failure, but has EXRESULT FAIL")
        tools_wrong = 'SYNTAX_FAILURE'

    params = {
        'file': file_with_path,
        'relfile': relfile,
        'todo': todo,
        'disabled': disabled,
        'tools_wrong': tools_wrong,
        'exresult': exresult,
    }

    TestParseParserTests.tests.append((params, exresult))
    return 0


def find_and_setup_test_profiles(profile_dir):
    """find all profiles in the given profile_dir, excluding
    - skippable files
    - include directories
    - files in the main directory (readme, todo etc.)
    """
    skipped = 0

    profile_dir = os.path.abspath(profile_dir)

    apparmor.profile_dir = profile_dir

    print('Searching for parser simple_tests... (this will take a while)')

    for root, dirs, files in os.walk(profile_dir):
        relpath = os.path.relpath(root, profile_dir)

        if relpath == '.':
            # include files are checked as part of the profiles that include them (also, they don't contain EXRESULT)
            dirs.remove('includes')
            dirs.remove('include_tests')
            dirs.remove('includes-preamble')

        for file in files:
            file_with_path = os.path.join(root, file)
            if not is_skippable_file(file) and relpath != '.':
                skipped += parse_test_profiles(file_with_path)

    if skipped:
        print('Skipping {} test profiles listed in skip_startswith.'.format(skipped))

    print('Running {} parser simple_tests...'.format(len(TestParseParserTests.tests)))


setup_aa(apparmor)
profile_dir = os.path.abspath('../../parser/tst/simple_tests/')
find_and_setup_test_profiles(profile_dir)

setup_all_loops(__name__)
if __name__ == '__main__':
    unittest.main(verbosity=1)



#!/usr/bin/python3
# ----------------------------------------------------------------------
#    Copyright (C) 2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import unittest
from collections import namedtuple

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.logparser import ReadLog
from apparmor.rule.dbus import DbusRule, DbusRuleset
from apparmor.translations import init_translation
from common_test import AATest, setup_all_loops

_ = init_translation()

exp = namedtuple(
    'exp', ('audit', 'allow_keyword', 'deny', 'comment', 'access', 'all_access',
            'bus', 'all_buses', 'path', 'all_paths', 'name', 'all_names',
            'interface', 'all_interfaces', 'member', 'all_members',
            'peername', 'all_peernames', 'peerlabel', 'all_peerlabels'))


# --- tests for single DbusRule --- #

class DbusTest(AATest):
    def _compare_obj(self, obj, expected):
        self.assertEqual(obj.allow_keyword, expected.allow_keyword)
        self.assertEqual(obj.audit, expected.audit)
        self.assertEqual(obj.deny, expected.deny)
        self.assertEqual(obj.comment, expected.comment)

        self.assertEqual(obj.access, expected.access)
        self._assertEqual_aare(obj.bus, expected.bus)
        self._assertEqual_aare(obj.path, expected.path)
        self._assertEqual_aare(obj.name, expected.name)
        self._assertEqual_aare(obj.interface, expected.interface)
        self._assertEqual_aare(obj.member, expected.member)
        self._assertEqual_aare(obj.peername, expected.peername)
        self._assertEqual_aare(obj.peerlabel, expected.peerlabel)

        self.assertEqual(obj.all_access, expected.all_access)
        self.assertEqual(obj.all_buses, expected.all_buses)
        self.assertEqual(obj.all_paths, expected.all_paths)
        self.assertEqual(obj.all_names, expected.all_names)
        self.assertEqual(obj.all_interfaces, expected.all_interfaces)
        self.assertEqual(obj.all_members, expected.all_members)
        self.assertEqual(obj.all_peernames, expected.all_peernames)
        self.assertEqual(obj.all_peerlabels, expected.all_peerlabels)

    def _assertEqual_aare(self, obj, expected):
        if obj:
            self.assertEqual(obj.regex, expected)
        else:
            self.assertEqual(obj, expected)


class DbusTestParse(DbusTest):
    tests = (
        # DbusRule object                                   audit  allow  deny   comment   access               all?   bus        all?   path        all?   name           all?    interface  all?  member  all?  peername  all?   peerlabel  all?
        ('dbus,',                                       exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus (   ),',                                 exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus ( , ),',                                 exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus send,',                                  exp(False, False, False, '',       {'send'},            False, None,      True,  None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus (send, receive),',                       exp(False, False, False, '',       {'send', 'receive'}, False, None,      True,  None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus send bus=session,',                      exp(False, False, False, '',       {'send'},            False, 'session', False, None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('deny dbus send bus="session", # cmt',         exp(False, False, True,  ' # cmt', {'send'},            False, 'session', False, None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('audit allow dbus peer=(label=foo),',          exp(True,  True,  False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, None,     True,  'foo',     False)),
        ('dbus bus=session path=/foo/bar,',             exp(False, False, False, '',       None,                True,  'session', False, '/foo/bar', False, None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus send bus=(session),',                    exp(False, False, False, '',       {'send'},            False, 'session', False, None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus name=(SomeService),',                    exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  'SomeService', False,  None,      True, None,   True, None,     True,  None,      True)),
        ('dbus send bus=session peer=(label="foo"),',   exp(False, False, False, '',       {'send'},            False, 'session', False, None,       True,  None,          True,   None,      True, None,   True, None,     True,  'foo',     False)),
        ('dbus send  bus = ( session ) , ',             exp(False, False, False, '',       {'send'},            False, 'session', False, None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus path=/foo,',                             exp(False, False, False, '',       None,                True,   None,     True,  '/foo',     False, None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus eavesdrop bus=session,',                 exp(False, False, False, '',       {'eavesdrop'},       False, 'session', False, None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus peer=(name=foo label=bar),',             exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo',    False, 'bar',     False)),
        ('dbus peer=(  name = foo    label = bar  ),',  exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo',    False, 'bar',     False)),
        ('dbus peer=(  name = foo ,  label = bar  ),',  exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo',    False, 'bar',     False)),
        ('dbus peer=(, name = foo ,  label = bar ,),',  exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo',    False, 'bar',     False)),
        ('dbus peer=(  name = foo,   label = bar  ),',  exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo,',   False, 'bar',     False)),  # XXX peername includes the comma
        ('dbus peer=(label=bar name=foo),',             exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo',    False, 'bar',     False)),
        ('dbus peer=(  label = bar   name = foo  ),',   exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo',    False, 'bar',     False)),
        ('dbus peer=(, label = bar , name = foo ,),',   exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo',    False, 'bar',     False)),
        ('dbus peer=(, label = bar,  name = foo  ),',   exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo',    False, 'bar,',    False)),  # XXX peerlabel includes the comma
        ('dbus peer=(  label = bar , name = foo  ),',   exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo',    False, 'bar',     False)),
        ('dbus peer=(  label = "bar"  name = "foo" ),', exp(False, False, False, '',       None,                True,  None,      True,  None,       True,  None,          True,   None,      True, None,   True, 'foo',    False, 'bar',     False)),
        ('dbus path=/foo/bar bus=session,',             exp(False, False, False, '',       None,                True,  'session', False, '/foo/bar', False, None,          True,   None,      True, None,   True, None,     True,  None,      True)),
        ('dbus bus=system path=/foo/bar bus=session,',  exp(False, False, False, '',       None,                True,  'session', False, '/foo/bar', False, None,          True,   None,      True, None,   True, None,     True,  None,      True)),  # XXX bus= specified twice, last one wins
        ('dbus send peer=(label="foo") bus=session,',   exp(False, False, False, '',       {'send'},            False, 'session', False, None,       True,  None,          True,   None,      True, None,   True, None,     True,  'foo',     False)),
        ('dbus bus=1 bus=2 bus=3 bus=4 bus=5 bus=6,',   exp(False, False, False, '',       None,                True,  '6',       False, None,       True,  None,          True,   None,      True, None,   True, None,     True,  None,      True)),  # XXX bus= specified multiple times, last one wins
    )

    def _run_test(self, rawrule, expected):
        self.assertTrue(DbusRule.match(rawrule))
        obj = DbusRule.create_instance(rawrule)
        self.assertEqual(rawrule.strip(), obj.raw_rule)
        self._compare_obj(obj, expected)


class DbusTestParseInvalid(DbusTest):
    tests = (
        ('dbus foo,',                                       AppArmorException),
        ('dbus foo bar,',                                   AppArmorException),
        ('dbus foo int,',                                   AppArmorException),
        ('dbus send bar,',                                  AppArmorException),
        ('dbus send receive,',                              AppArmorException),
        ('dbus peer=,',                                     AppArmorException),
        ('dbus peer=(label=foo) path=,',                    AppArmorException),
        ('dbus (invalid),',                                 AppArmorException),
        ('dbus peer=,',                                     AppArmorException),
        ('dbus bus=session bind bus=system,',               AppArmorException),
        ('dbus bus=1 bus=2 bus=3 bus=4 bus=5 bus=6 bus=7,', AppArmorException),
    )

    def _run_test(self, rawrule, expected):
        self.assertTrue(DbusRule.match(rawrule))  # the above invalid rules still match the main regex!
        with self.assertRaises(expected):
            DbusRule.create_instance(rawrule)


class DbusTestParseFromLog(DbusTest):
    def test_dbus_from_log(self):
        parser = ReadLog('', '', '')
        event = 'type=USER_AVC msg=audit(1375323372.644:157): pid=363 uid=102 auid=4294967295 ses=4294967295  msg=\'apparmor="DENIED" operation="dbus_method_call"  bus="system" name="org.freedesktop.DBus" path="/org/freedesktop/DBus" interface="org.freedesktop.DBus" member="Hello" mask="send" pid=2833 profile="/tmp/apparmor-2.8.0/tests/regression/apparmor/dbus_service" peer_profile="unconfined"  exe="/bin/dbus-daemon" sauid=102 hostname=? addr=? terminal=?\''

        parsed_event = parser.parse_event(event)

        self.assertEqual(parsed_event, {
            'request_mask': None,
            'denied_mask': 'send',
            'error_code': 0,
            'magic_token': 0,
            'parent': 0,
            'profile': '/tmp/apparmor-2.8.0/tests/regression/apparmor/dbus_service',
            'bus': 'system',
            'peer_profile': 'unconfined',
            'operation': 'dbus_method_call',
            'resource': None,
            'info': None,
            'aamode': 'REJECTING',
            'time': 1375323372,
            'active_hat': None,
            'pid': 2833,
            'task': 0,
            'attr': None,
            'name2': None,
            'name': 'org.freedesktop.DBus',
            'path': '/org/freedesktop/DBus',
            'interface': 'org.freedesktop.DBus',
            'member': 'Hello',
            'family': None,
            'protocol': None,
            'sock_type': None,
            'class': None,
        })

        # # XXX send rules must not contain name conditional, but the log event includes it - how should we handle this in logparser.py?
        #
        # #                            access                       bus                  path                  name                  interface                  member                  peername        peerlabel
        # obj = DbusRule(parsed_event['denied_mask'], parsed_event['bus'], parsed_event['path'], parsed_event['name'], parsed_event['interface'], parsed_event['member'], parsed_event['peer_profile'], DbusRule.ALL, log_event=parsed_event)
        #
        # # DbusRule     audit  allow  deny   comment  access    all?   bus       all?   path                     all?   name                    all?   interface               all?    member  all?   peername      all?   peerlabel  all?
        # expected = exp(False, False, False, '',      {'send'}, False, 'system', False, '/org/freedesktop/DBus', False, 'org.freedesktop.DBus', False, 'org.freedesktop.DBus', False, 'Hello', False, 'unconfined', False, None,      True)
        #
        # self._compare_obj(obj, expected)
        #
        # self.assertEqual(obj.get_raw(1), '  dbus send bus=system path=/org/freedesktop/DBus name=org.freedesktop.DBus member=Hello peer=(name=unconfined),')


class DbusFromInit(DbusTest):
    tests = (
        #         access  bus        path          name          interface     member        peername      peerlabel   audit=, deny=, allow_keyword, comment=, log_event)
        (DbusRule('send', 'session', DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),
         #   audit  allow  deny   comment  access    all?   bus        all?   path  all?  name  all?  interface all?  member all?  peername all?  peerlabel all?
         exp(False, False, False, '',      {'send'}, False, 'session', False, None, True, None, True, None,     True, None,  True, None,    True, None,     True)),

        #         access               bus        path          name          interface     member        peername      peerlabel   audit=, deny=, allow_keyword, comment=, log_event)
        (DbusRule(('send', 'receive'), 'session', DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),
         #   audit  allow  deny   comment access               all?   bus        all?   path  all?  name  all?  interface all?  member all?  peername all?  peerlabel all?
         exp(False, False, False, '',     {'send', 'receive'}, False, 'session', False, None, True, None, True, None,     True, None,  True, None,    True, None,     True)),

        #         access        bus           path          name          interface    member      peername      peerlabel   audit=, deny=, allow_keyword, comment=, log_event)
        (DbusRule(DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, '/int/face', '/mem/ber', '/peer/name', '/peer/label'),
         #   audit  allow  deny   comment access  all?  bus   all?  path  all?  name  all?  interface    all?   member      all?   peername      all?   peerlabel      all?
         exp(False, False, False, '',     None,   True, None, True, None, True, None, True, '/int/face', False, '/mem/ber', False, '/peer/name', False, '/peer/label', False)),
    )

    def _run_test(self, obj, expected):
        self._compare_obj(obj, expected)


class InvalidDbusInit(AATest):
    tests = (
        #  access          bus          path           name           interface     member        peername      peerlabel       expected exception

        # empty fields
        (('',              'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        (((),              'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    '',          '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '',            'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '/org/test',   '',            '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '/org/test',   'com.aa.test', '',           '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '/org/test',   'com.aa.test', '/int/face',  '',           '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '',           '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', ''),            AppArmorBug),

        # whitespace fields
        (('  ',            'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    '  ',        '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '  ',          'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '/org/test',   '  ',          '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '/org/test',   'com.aa.test', '  ',         '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '/org/test',   'com.aa.test', '/int/face',  '  ',         '/peer/name', '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '  ',         '/peer/label'), AppArmorBug),
        ((DbusRule.ALL,    'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '  '),          AppArmorBug),

        # wrong type - dict()
        ((dict(),          'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        dict(),      '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        'session',   dict(),        'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        'session',   '/org/test',   dict(),        '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        'session',   '/org/test',   'com.aa.test', dict(),       '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        'session',   '/org/test',   'com.aa.test', '/int/face',  dict(),       '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   dict(),       '/peer/label'), AppArmorBug),
        ((('send'),        'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', dict()),        AppArmorBug),


        # wrong type - None
        ((None,            'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        (((None),          'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        None,        '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        'session',   None,          'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        'session',   '/org/test',   None,          '/int/face',  '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        'session',   '/org/test',   'com.aa.test', None,         '/mem/ber',   '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        'session',   '/org/test',   'com.aa.test', '/int/face',  None,         '/peer/name', '/peer/label'), AppArmorBug),
        ((('send'),        'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   None,         '/peer/label'), AppArmorBug),
        ((('send'),        'session',   '/org/test',   'com.aa.test', '/int/face',  '/mem/ber',   '/peer/name', None),          AppArmorBug),

        # bind conflicts with path, interface, member, peer name and peer label
        ((('bind'),        DbusRule.ALL, '/org/test',  DbusRule.ALL,  DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),  AppArmorException),
        ((('bind'),        DbusRule.ALL, DbusRule.ALL, DbusRule.ALL,  '/int/face',  DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),  AppArmorException),
        ((('bind'),        DbusRule.ALL, DbusRule.ALL, DbusRule.ALL,  DbusRule.ALL, '/mem/ber',   DbusRule.ALL, DbusRule.ALL),  AppArmorException),
        ((('bind'),        DbusRule.ALL, DbusRule.ALL, DbusRule.ALL,  DbusRule.ALL, DbusRule.ALL, '/peer/name', DbusRule.ALL),  AppArmorException),
        ((('bind'),        DbusRule.ALL, DbusRule.ALL, DbusRule.ALL,  DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, '/peer/label'), AppArmorException),

        # eavesdrop conflcts with path, name, interface, member, peer name and peer label
        ((('eavesdrop'),   DbusRule.ALL, '/org/test',  DbusRule.ALL,  DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),  AppArmorException),
        ((('eavesdrop'),   DbusRule.ALL, DbusRule.ALL, 'com.aa.test', DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),  AppArmorException),
        ((('eavesdrop'),   DbusRule.ALL, DbusRule.ALL, DbusRule.ALL,  '/int/face',  DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),  AppArmorException),
        ((('eavesdrop'),   DbusRule.ALL, DbusRule.ALL, DbusRule.ALL,  DbusRule.ALL, '/mem/ber',   DbusRule.ALL, DbusRule.ALL),  AppArmorException),
        ((('eavesdrop'),   DbusRule.ALL, DbusRule.ALL, DbusRule.ALL,  DbusRule.ALL, DbusRule.ALL, '/peer/name', DbusRule.ALL),  AppArmorException),
        ((('eavesdrop'),   DbusRule.ALL, DbusRule.ALL, DbusRule.ALL,  DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, '/peer/label'), AppArmorException),

        # send and receive conflict with name
        ((('send'),        DbusRule.ALL, DbusRule.ALL, 'com.aa.test', DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),  AppArmorException),
        ((('receive'),     DbusRule.ALL, DbusRule.ALL, 'com.aa.test', DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),  AppArmorException),

        # misc
        ((DbusRule.ALL,    DbusRule.ALL, 'foo/bar',    DbusRule.ALL,  DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),  AppArmorException),  # path doesn't start with /
        ((('foo'),         DbusRule.ALL, DbusRule.ALL, DbusRule.ALL,  DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),  AppArmorException),  # invalid access keyword
        ((('foo', 'send'), DbusRule.ALL, DbusRule.ALL, DbusRule.ALL,  DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL),  AppArmorException),  # valid + invalid access keyword
    )

    def _run_test(self, params, expected):
        with self.assertRaises(expected):
            DbusRule(*params)

    def test_missing_params_1(self):
        with self.assertRaises(TypeError):
            DbusRule('send')

    def test_missing_params_2(self):
        with self.assertRaises(TypeError):
            DbusRule(('send'), 'session')

    def test_missing_params_3(self):
        with self.assertRaises(TypeError):
            DbusRule(('send'), 'session', '/org/test')

    def test_missing_params_4(self):
        with self.assertRaises(TypeError):
            DbusRule(('send'), 'session', '/org/test', 'com.aa.test')

    def test_missing_params_5(self):
        with self.assertRaises(TypeError):
            DbusRule(('send'), 'session', '/org/test', 'com.aa.test', '/int/face')

    def test_missing_params_6(self):
        with self.assertRaises(TypeError):
            DbusRule(('send'), 'session', '/org/test', 'com.aa.test', '/int/face', '/mem/ber')

    def test_missing_params_7(self):
        with self.assertRaises(TypeError):
            DbusRule(('send'), 'session', '/org/test', 'com.aa.test', '/int/face', '/mem/ber', '/peer/name')


class InvalidDbusTest(AATest):
    def _check_invalid_rawrule(self, rawrule):
        obj = None
        self.assertFalse(DbusRule.match(rawrule))
        with self.assertRaises(AppArmorException):
            obj = DbusRule.create_instance(rawrule)

        self.assertIsNone(obj, 'DbusRule handed back an object unexpectedly')

    def test_invalid_dbus_missing_comma(self):
        self._check_invalid_rawrule('dbus')  # missing comma

    def test_invalid_non_DbusRule(self):
        self._check_invalid_rawrule('signal,')  # not a dbus rule

    def test_empty_data_1(self):
        #              access        bus        path         name          interface    member      peername      peerlabel
        obj = DbusRule(('send'),     'session', '/org/test', DbusRule.ALL, '/int/face', '/mem/ber', '/peer/name', '/peer/label')
        obj.access = ''
        # no access set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)

    def test_empty_data_2(self):
        obj = DbusRule(('send'),     'session', '/org/test', DbusRule.ALL, '/int/face', '/mem/ber', '/peer/name', '/peer/label')
        obj.bus = ''
        # no bus set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)

    def test_empty_data_3(self):
        obj = DbusRule(('send'),     'session', '/org/test', DbusRule.ALL, '/int/face', '/mem/ber', '/peer/name', '/peer/label')
        obj.path = ''
        # no path set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)

    def test_empty_data_4(self):
        obj = DbusRule(DbusRule.ALL, 'session', '/org/test', 'com.aa.test', '/int/face', '/mem/ber', '/peer/name', '/peer/label')
        obj.name = ''
        # no name set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)

    def test_empty_data_5(self):
        obj = DbusRule(('send'),     'session', '/org/test', DbusRule.ALL, '/int/face', '/mem/ber', '/peer/name', '/peer/label')
        obj.interface = ''
        # no interface set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)

    def test_empty_data_6(self):
        obj = DbusRule(('send'),     'session', '/org/test', DbusRule.ALL, '/int/face', '/mem/ber', '/peer/name', '/peer/label')
        obj.member = ''
        # no member set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)

    def test_empty_data_7(self):
        obj = DbusRule(('send'),     'session', '/org/test', DbusRule.ALL, '/int/face', '/mem/ber', '/peer/name', '/peer/label')
        obj.peername = ''
        # no peername set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)

    def test_empty_data_8(self):
        obj = DbusRule(('send'),     'session', '/org/test', DbusRule.ALL, '/int/face', '/mem/ber', '/peer/name', '/peer/label')
        obj.peerlabel = ''
        # no peerlabel set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)

    def test_invalid_event_access(self):
        parser = ReadLog('', '', '')
        event = 'type=USER_AVC msg=audit(1375323372.644:157): pid=363 uid=102 auid=4294967295 ses=4294967295  msg=\'apparmor="DENIED" operation="dbus_method_call"  bus="system" name="org.freedesktop.DBus" path="/org/freedesktop/DBus" interface="org.freedesktop.DBus" member="Hello" mask="invalid_access" pid=2833 profile="/tmp/apparmor-2.8.0/tests/regression/apparmor/dbus_service" peer_profile="unconfined"  exe="/bin/dbus-daemon" sauid=102 hostname=? addr=? terminal=?\''
        ev = parser.parse_event(event)
        with self.assertRaises(AppArmorBug):
            DbusRule.create_from_ev(ev)


class WriteDbusTest(AATest):
    def _run_test(self, rawrule, expected):
        self.assertTrue(DbusRule.match(rawrule), 'DbusRule.match() failed')
        obj = DbusRule.create_instance(rawrule)
        clean = obj.get_clean()
        raw = obj.get_raw()

        self.assertEqual(expected.strip(), clean, 'unexpected clean rule')
        self.assertEqual(rawrule.strip(), raw, 'unexpected raw rule')

    tests = (
        #  raw rule                                                          clean rule
        ('     dbus         ,    # foo        ',                             'dbus, # foo'),
        ('    audit     dbus send,',                                         'audit dbus send,'),
        ('    audit     dbus (send  ),',                                     'audit dbus send,'),
        ('    audit     dbus (send , receive ),',                            'audit dbus (receive send),'),
        ('   deny dbus         send      bus=session,# foo bar',             'deny dbus send bus=session, # foo bar'),
        ('   deny dbus         send      bus=(session),  ',                  'deny dbus send bus=session,'),
        ('   deny dbus         send      peer=(name=unconfined label=foo),', 'deny dbus send peer=(name=unconfined label=foo),'),
        ('   deny dbus         send      interface = ( foo ),',              'deny dbus send interface=foo,'),
        ('   deny dbus         send      ,# foo bar',                        'deny dbus send, # foo bar'),
        ('   allow dbus             peer=(label=foo)    ,# foo bar',         'allow dbus peer=(label=foo), # foo bar'),
        ('dbus,',                                                            'dbus,'),
        ('dbus (receive),',                                                  'dbus receive,'),
        ('dbus (send),',                                                     'dbus send,'),
        ('dbus (send receive),',                                             'dbus (receive send),'),
        ('dbus receive,',                                                    'dbus receive,'),
        ('dbus eavesdrop,',                                                  'dbus eavesdrop,'),
        ('dbus bind bus = foo name = bar,',                                  'dbus bind bus=foo name=bar,'),
        ('dbus send peer=(  label = /foo ) ,',                               'dbus send peer=(label=/foo),'),
        ('dbus (receive) member=baz,',                                       'dbus receive member=baz,'),
        ('dbus send path = /foo,',                                           'dbus send path=/foo,'),
        ('dbus receive peer=(label=foo),',                                   'dbus receive peer=(label=foo),'),
        ('dbus (send receive) peer=(name=/usr/bin/bar),',                    'dbus (receive send) peer=(name=/usr/bin/bar),'),
        ('dbus (,  receive  ,,, send  ,) interface=/sbin/baz,',              'dbus (receive send) interface=/sbin/baz,'),  # XXX leading and trailing ',' inside (...) causes error
        ('  priority=123 deny dbus         send      interface = ( foo ),',  'priority=123 deny dbus send interface=foo,'),
        ('  priority=0   deny dbus         send      interface = ( foo ),',  'priority=0 deny dbus send interface=foo,'),
        ('  priority=-72 deny dbus         send      interface = ( foo ),',  'priority=-72 deny dbus send interface=foo,'),
        ('  priority=+834 deny dbus         send      interface = ( foo ),', 'priority=834 deny dbus send interface=foo,'),
        # XXX add more complex rules
    )

    def test_write_manually_1(self):
        #              access    bus        path         name          interface    member      peername      peerlabel
        obj = DbusRule(('send'), 'session', '/org/test', DbusRule.ALL, '/int/face', '/mem/ber', '/peer/name', '/peer/label', allow_keyword=True)

        expected = '    allow dbus send bus=session path=/org/test interface=/int/face member=/mem/ber peer=(name=/peer/name label=/peer/label),'

        self.assertEqual(expected, obj.get_clean(2), 'unexpected clean rule')
        self.assertEqual(expected, obj.get_raw(2), 'unexpected raw rule')

    def test_write_manually_2(self):
        #              access               bus           path         name          interface     member      peername      peerlabel
        obj = DbusRule(('send', 'receive'), DbusRule.ALL, '/org/test', DbusRule.ALL, DbusRule.ALL, '/mem/ber', '/peer/name', DbusRule.ALL, allow_keyword=True)

        expected = '    allow dbus (receive send) path=/org/test member=/mem/ber peer=(name=/peer/name),'

        self.assertEqual(expected, obj.get_clean(2), 'unexpected clean rule')
        self.assertEqual(expected, obj.get_raw(2), 'unexpected raw rule')


class DbusCoveredTest(AATest):
    def _run_test(self, param, expected):
        obj = DbusRule.create_instance(self.rule)
        check_obj = DbusRule.create_instance(param)

        self.assertTrue(DbusRule.match(param))

        self.assertEqual(obj.is_equal(check_obj), expected[0], 'Mismatch in is_equal, expected {}'.format(expected[0]))
        self.assertEqual(obj.is_equal(check_obj, True), expected[1], 'Mismatch in is_equal/strict, expected {}'.format(expected[1]))

        self.assertEqual(obj.is_covered(check_obj), expected[2], 'Mismatch in is_covered, expected {}'.format(expected[2]))
        self.assertEqual(obj.is_covered(check_obj, True, True), expected[3], 'Mismatch in is_covered/exact, expected {}'.format(expected[3]))


class DbusCoveredTest_01(DbusCoveredTest):
    rule = 'dbus send,'

    tests = (
        #   rule                           equal  strict equal  covered  covered exact
        ('dbus,',                         (False, False,        False,   False)),
        ('dbus send,',                    (True,  True,         True,    True)),
        ('dbus send member=unconfined,',  (False, False,        True,    True)),
        ('dbus send, # comment',          (True,  False,        True,    True)),
        ('allow dbus send,',              (True,  False,        True,    True)),
        ('dbus     send,',                (True,  False,        True,    True)),
        ('dbus send bus=session,',        (False, False,        True,    True)),
        ('dbus send member=(label=foo),', (False, False,        True,    True)),
        ('audit dbus send,',              (False, False,        False,   False)),
        ('audit dbus,',                   (False, False,        False,   False)),
        ('dbus receive,',                 (False, False,        False,   False)),
        ('dbus member=(label=foo),',      (False, False,        False,   False)),
        ('audit deny dbus send,',         (False, False,        False,   False)),
        ('deny dbus send,',               (False, False,        False,   False)),
    )


class DbusCoveredTest_02(DbusCoveredTest):
    rule = 'audit dbus send,'

    tests = (
        #   rule                          equal  strict equal  covered  covered exact
        ('      dbus send,',             (False, False,        True,    False)),
        ('audit dbus send,',             (True,  True,         True,    True)),
        ('      dbus send bus=session,', (False, False,        True,    False)),
        ('audit dbus send bus=session,', (False, False,        True,    True)),
        ('      dbus,',                  (False, False,        False,   False)),
        ('audit dbus,',                  (False, False,        False,   False)),
        ('dbus receive,',                (False, False,        False,   False)),
    )


class DbusCoveredTest_03(DbusCoveredTest):
    rule = 'dbus send bus=session,'

    tests = (
        #   rule                                 equal  strict equal  covered  covered exact
        ('      dbus send bus=session,',        (True,  True,         True,    True)),
        ('allow dbus send bus=session,',        (True,  False,        True,    True)),
        ('      dbus send,',                    (False, False,        False,   False)),
        ('      dbus,',                         (False, False,        False,   False)),
        ('      dbus send member=(label=foo),', (False, False,        False,   False)),
        ('audit dbus,',                         (False, False,        False,   False)),
        ('audit dbus send bus=session,',        (False, False,        False,   False)),
        ('audit dbus bus=session,',             (False, False,        False,   False)),
        ('      dbus send,',                    (False, False,        False,   False)),
        ('      dbus,',                         (False, False,        False,   False)),
    )


class DbusCoveredTest_04(DbusCoveredTest):
    rule = 'dbus,'

    tests = (
        #   rule                             equal  strict equal  covered  covered exact
        ('      dbus,',                     (True,  True,         True,    True)),
        ('allow dbus,',                     (True,  False,        True,    True)),
        ('      dbus send,',                (False, False,        True,    True)),
        ('      dbus receive bus=session,', (False, False,        True,    True)),
        ('      dbus member=(label=foo),',  (False, False,        True,    True)),
        ('      dbus send bus=session,',    (False, False,        True,    True)),
        ('audit dbus,',                     (False, False,        False,   False)),
        ('deny  dbus,',                     (False, False,        False,   False)),
    )


class DbusCoveredTest_05(DbusCoveredTest):
    rule = 'deny dbus send,'

    tests = (
        #   rule                     equal  strict equal  covered  covered exact
        ('      deny dbus send,',    (True,  True,       True,    True)),
        ('audit deny dbus send,',    (False, False,      False,   False)),
        ('           dbus send,',    (False, False,      False,   False)),  # XXX should covered be true here?
        ('      deny dbus receive,', (False, False,      False,   False)),
        ('      deny dbus,',         (False, False,      False,   False)),
    )


class DbusCoveredTest_06(DbusCoveredTest):
    rule = 'dbus send peer=(name=unconfined),'

    tests = (
        #   rule                                        equal  strict equal  covered  covered exact
        ('dbus,',                                       (False, False,      False,   False)),
        ('dbus send,',                                  (False, False,      False,   False)),
        ('dbus send peer=(name=unconfined),',           (True,  True,       True,    True)),
        ('dbus peer=(name=unconfined),',                (False, False,      False,   False)),
        ('dbus send, # comment',                        (False, False,      False,   False)),
        ('allow dbus send,',                            (False, False,      False,   False)),
        ('allow dbus send peer=(name=unconfined),',     (True,  False,      True,    True)),
        ('allow dbus send peer=(name=/foo/bar),',       (False, False,      False,   False)),
        ('allow dbus send peer=(name=/**),',            (False, False,      False,   False)),
        ('allow dbus send peer=(name=**),',             (False, False,      False,   False)),
        ('dbus    send,',                               (False, False,      False,   False)),
        ('dbus    send peer=(name=unconfined),',        (True,  False,      True,    True)),
        ('dbus send bus=session,',                      (False, False,      False,   False)),
        ('dbus send peer=(name=unconfined label=foo),', (False, False,      True,    True)),
        ('audit dbus send peer=(name=unconfined),',     (False, False,      False,   False)),
        ('audit dbus,',                                 (False, False,      False,   False)),
        ('dbus receive,',                               (False, False,      False,   False)),
        ('dbus peer=(label=foo),',                      (False, False,      False,   False)),
        ('audit deny dbus send,',                       (False, False,      False,   False)),
        ('deny dbus send,',                             (False, False,      False,   False)),
    )


class DbusCoveredTest_07(DbusCoveredTest):
    rule = 'dbus send peer=(label=unconfined),'

    tests = (
        #   rule                                        equal  strict equal  covered  covered exact
        ('dbus,',                                       (False, False,      False,   False)),
        ('dbus send,',                                  (False, False,      False,   False)),
        ('dbus send peer=(label=unconfined),',          (True,  True,       True,    True)),
        ('dbus peer=(label=unconfined),',               (False, False,      False,   False)),
        ('dbus send, # comment',                        (False, False,      False,   False)),
        ('allow dbus send,',                            (False, False,      False,   False)),
        ('allow dbus send peer=(label=unconfined),',    (True,  False,      True,    True)),
        ('allow dbus send peer=(label=/foo/bar),',      (False, False,      False,   False)),
        ('allow dbus send peer=(label=/**),',           (False, False,      False,   False)),
        ('allow dbus send peer=(label=**),',            (False, False,      False,   False)),
        ('dbus    send,',                               (False, False,      False,   False)),
        ('dbus    send peer=(label=unconfined),',       (True,  False,      True,    True)),
        ('dbus send bus=session,',                      (False, False,      False,   False)),
        ('dbus send peer=(label=unconfined name=foo),', (False, False,      True,    True)),
        ('audit dbus send peer=(label=unconfined),',    (False, False,      False,   False)),
        ('audit dbus,',                                 (False, False,      False,   False)),
        ('dbus receive,',                               (False, False,      False,   False)),
        ('dbus peer=(label=foo),',                      (False, False,      False,   False)),
        ('audit deny dbus send,',                       (False, False,      False,   False)),
        ('deny dbus send,',                             (False, False,      False,   False)),
    )


class DbusCoveredTest_08(DbusCoveredTest):
    rule = 'dbus send path=/foo/bar,'

    tests = (
        #   rule                                      equal  strict equal  covered  covered exact
        ('dbus,',                                     (False, False,      False,   False)),
        ('dbus send,',                                (False, False,      False,   False)),
        ('dbus send path=/foo/bar,',                  (True,  True,       True,    True)),
        ('dbus send path=/foo/*,',                    (False, False,      False,   False)),
        ('dbus send path=/**,',                       (False, False,      False,   False)),
        ('dbus send path=/what/*,',                   (False, False,      False,   False)),
        ('dbus path=/foo/bar,',                       (False, False,      False,   False)),
        ('dbus send, # comment',                      (False, False,      False,   False)),
        ('allow dbus send,',                          (False, False,      False,   False)),
        ('allow dbus send path=/foo/bar,',            (True,  False,      True,    True)),
        ('dbus    send,',                             (False, False,      False,   False)),
        ('dbus    send path=/foo/bar,',               (True,  False,      True,    True)),
        ('dbus    send path=/what/ever,',             (False, False,      False,   False)),
        ('dbus send bus=session,',                    (False, False,      False,   False)),
        ('dbus send path=/foo/bar peer=(label=foo),', (False, False,      True,    True)),
        ('audit dbus send path=/foo/bar,',            (False, False,      False,   False)),
        ('audit dbus,',                               (False, False,      False,   False)),
        ('dbus receive,',                             (False, False,      False,   False)),
        ('dbus peer=(label=foo),',                    (False, False,      False,   False)),
        ('audit deny dbus send,',                     (False, False,      False,   False)),
        ('deny dbus send,',                           (False, False,      False,   False)),
    )


class DbusCoveredTest_09(DbusCoveredTest):
    rule = 'dbus send member=**,'

    tests = (
        #   rule                                        equal  strict equal  covered  covered exact
        ('dbus,',                                       (False, False,      False,   False)),
        ('dbus send,',                                  (False, False,      False,   False)),
        ('dbus send member=/foo/bar,',                  (False, False,      True,    True)),
        ('dbus send member=/foo/*,',                    (False, False,      False,   False)),  # TODO: wildcard vs. wildcard never matches in is_covered_aare()
        ('dbus send member=/**,',                       (False, False,      False,   False)),  # TODO: wildcard vs. wildcard never matches in is_covered_aare()
        ('dbus send member=/what/*,',                   (False, False,      False,   False)),  # TODO: wildcard vs. wildcard never matches in is_covered_aare()
        ('dbus member=/foo/bar,',                       (False, False,      False,   False)),
        ('dbus send, # comment',                        (False, False,      False,   False)),
        ('allow dbus send,',                            (False, False,      False,   False)),
        ('allow dbus send member=/foo/bar,',            (False, False,      True,    True)),
        ('dbus    send,',                               (False, False,      False,   False)),
        ('dbus    send member=/foo/bar,',               (False, False,      True,    True)),
        ('dbus    send member=/what/ever,',             (False, False,      True,    True)),
        ('dbus send bus=session,',                      (False, False,      False,   False)),
        ('dbus send member=/foo/bar peer=(label=foo),', (False, False,      True,    True)),
        ('audit dbus send member=/foo/bar,',            (False, False,      False,   False)),
        ('audit dbus,',                                 (False, False,      False,   False)),
        ('dbus receive,',                               (False, False,      False,   False)),
        ('dbus member=(label=foo),',                    (False, False,      False,   False)),
        ('audit deny dbus send,',                       (False, False,      False,   False)),
        ('deny dbus send,',                             (False, False,      False,   False)),
    )


class DbusCoveredTest_10(DbusCoveredTest):
    rule = 'dbus (send, receive) interface=foo,'

    tests = (
        #   rule                                        equal  strict equal  covered  covered exact
        ('dbus,',                                       (False, False,      False,    False)),
        ('dbus send,',                                  (False, False,      False,    False)),
        ('dbus send interface=foo,',                    (False, False,      True,     True)),
        ('dbus receive bus=session interface=foo,',     (False, False,      True,     True)),
        ('dbus (receive,send) interface=foo,',          (True,  False,      True,     True)),
        ('dbus (receive,send),',                        (False, False,      False,    False)),
        ('dbus send bus=session,',                      (False, False,      False,    False)),
        ('dbus send member=/foo/bar,',                  (False, False,      False,    False)),
        ('dbus send member=/foo/*,',                    (False, False,      False,    False)),
        ('dbus send member=/**,',                       (False, False,      False,    False)),
        ('dbus send member=/what/*,',                   (False, False,      False,    False)),
        ('dbus member=/foo/bar,',                       (False, False,      False,    False)),
        ('dbus send, # comment',                        (False, False,      False,    False)),
        ('allow dbus send,',                            (False, False,      False,    False)),
        ('allow dbus send member=/foo/bar,',            (False, False,      False,    False)),
        ('dbus    send,',                               (False, False,      False,    False)),
        ('dbus    send member=/foo/bar,',               (False, False,      False,    False)),
        ('dbus    send member=/what/ever,',             (False, False,      False,    False)),
        ('dbus send bus=session,',                      (False, False,      False,    False)),
        ('dbus send bus=session interface=foo,',        (False, False,      True,     True)),
        ('dbus send member=/foo/bar peer=(label=foo),', (False, False,      False,    False)),
        ('audit dbus send member=/foo/bar,',            (False, False,      False,    False)),
        ('audit dbus,',                                 (False, False,      False,    False)),
        ('dbus receive,',                               (False, False,      False,    False)),
        ('dbus peer=(label=foo),',                      (False, False,      False,    False)),
        ('audit deny dbus send,',                       (False, False,      False,    False)),
        ('deny dbus send,',                             (False, False,      False,    False)),
    )


class DbusCoveredTest_11(DbusCoveredTest):
    rule = 'dbus name=/foo/bar,'

    tests = (
        #   rule                                 equal  strict equal  covered  covered exact
        ('dbus,',                                (False, False,     False,   False)),
        ('dbus name=/foo/bar,',                  (True,  True,      True,    True)),
        ('dbus name=/foo/*,',                    (False, False,     False,   False)),
        ('dbus name=/**,',                       (False, False,     False,   False)),
        ('dbus name=/what/*,',                   (False, False,     False,   False)),
        ('dbus, # comment',                      (False, False,     False,   False)),
        ('allow dbus,',                          (False, False,     False,   False)),
        ('allow dbus name=/foo/bar,',            (True,  False,     True,    True)),
        ('dbus   ,',                             (False, False,     False,   False)),
        ('dbus    name=/foo/bar,',               (True,  False,     True,    True)),
        ('dbus    name=/what/ever,',             (False, False,     False,   False)),
        ('dbus bus=session,',                    (False, False,     False,   False)),
        ('dbus name=/foo/bar peer=(label=foo),', (False, False,     True,    True)),
        ('audit dbus name=/foo/bar,',            (False, False,     False,   False)),
        ('audit dbus,',                          (False, False,     False,   False)),
        ('dbus receive,',                        (False, False,     False,   False)),
        ('dbus peer=(label=foo),',               (False, False,     False,   False)),
        ('audit deny dbus,',                     (False, False,     False,   False)),
        ('deny dbus,',                           (False, False,     False,   False)),
    )


class DbusCoveredTest_Priority(DbusCoveredTest):
    rule = 'dbus send,'

    tests = (
        #   rule                         equal  strict equal  covered  covered exact
        ('priority=1 dbus send,',        (False, False,       False,   False)),
        ('priority=-1 dbus send,',       (False, False,       True,    True)),
        ('priority=0 dbus send,',        (True,  False,       True,    True)),
    )


class DbusCoveredTest_Invalid(AATest):
    def AASetup(self):
        #                        access               bus        path         name          interface    member        peername      peerlabel
        self.obj     = DbusRule(('send', 'receive'), 'session', '/org/test', DbusRule.ALL, '/int/face', DbusRule.ALL, '/peer/name', '/peer/label', allow_keyword=True)  # noqa: E221
        self.testobj = DbusRule(('send'),            'session', '/org/test', DbusRule.ALL, '/int/face', '/mem/ber',   '/peer/name', '/peer/label', allow_keyword=True)

    def test_borked_obj_is_covered_1(self):
        self.testobj.access = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_borked_obj_is_covered_2(self):
        self.testobj.bus = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_borked_obj_is_covered_3(self):
        self.testobj.path = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_borked_obj_is_covered_4(self):
        # we need a different 'victim' because dbus send doesn't allow the name conditional we want to test here
        self.obj     = DbusRule(('bind'), 'session',  DbusRule.ALL, '/name', DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, allow_keyword=True)  # noqa: E221
        self.testobj = DbusRule(('bind'), 'session',  DbusRule.ALL, '/name', DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, DbusRule.ALL, allow_keyword=True)
        self.testobj.name = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_borked_obj_is_covered_5(self):
        self.testobj.interface = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_borked_obj_is_covered_6(self):
        self.testobj.member = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_borked_obj_is_covered_7(self):
        self.testobj.peername = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_borked_obj_is_covered_8(self):
        self.testobj.peerlabel = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_invalid_is_covered(self):
        raw_rule = 'dbus send,'

        class SomeOtherClass(DbusRule):
            pass

        obj = DbusRule.create_instance(raw_rule)
        testobj = SomeOtherClass.create_instance(raw_rule)  # different type
        with self.assertRaises(AppArmorBug):
            obj.is_covered(testobj)

    def test_invalid_is_equal(self):
        raw_rule = 'dbus send,'

        class SomeOtherClass(DbusRule):
            pass

        obj = DbusRule.create_instance(raw_rule)
        testobj = SomeOtherClass.create_instance(raw_rule)  # different type
        with self.assertRaises(AppArmorBug):
            obj.is_equal(testobj)


class DbusLogprofHeaderTest(AATest):
    tests = (
        ('dbus,',                        [                              _('Access mode'), _('ALL'),       _('Bus'), _('ALL'),  _('Path'), _('ALL'), _('Name'), _('ALL'),    _('Interface'), _('ALL'),  _('Member'), _('ALL'), _('Peer name'), _('ALL'),    _('Peer label'), _('ALL')]),  # noqa: E201
        ('dbus (send receive),',         [                              _('Access mode'), 'receive send', _('Bus'), _('ALL'),  _('Path'), _('ALL'), _('Name'), _('ALL'),    _('Interface'), _('ALL'),  _('Member'), _('ALL'), _('Peer name'), _('ALL'),    _('Peer label'), _('ALL')]),  # noqa: E201
        ('dbus send bus=session,',       [                              _('Access mode'), 'send',         _('Bus'), 'session', _('Path'), _('ALL'), _('Name'), _('ALL'),    _('Interface'), _('ALL'),  _('Member'), _('ALL'), _('Peer name'), _('ALL'),    _('Peer label'), _('ALL')]),  # noqa: E201
        ('deny dbus,',                   [_('Qualifier'), 'deny',       _('Access mode'), _('ALL'),       _('Bus'), _('ALL'),  _('Path'), _('ALL'), _('Name'), _('ALL'),    _('Interface'), _('ALL'),  _('Member'), _('ALL'), _('Peer name'), _('ALL'),    _('Peer label'), _('ALL')]),
        ('allow dbus send,',             [_('Qualifier'), 'allow',      _('Access mode'), 'send',         _('Bus'), _('ALL'),  _('Path'), _('ALL'), _('Name'), _('ALL'),    _('Interface'), _('ALL'),  _('Member'), _('ALL'), _('Peer name'), _('ALL'),    _('Peer label'), _('ALL')]),
        ('audit dbus send bus=session,', [_('Qualifier'), 'audit',      _('Access mode'), 'send',         _('Bus'), 'session', _('Path'), _('ALL'), _('Name'), _('ALL'),    _('Interface'), _('ALL'),  _('Member'), _('ALL'), _('Peer name'), _('ALL'),    _('Peer label'), _('ALL')]),
        ('audit deny dbus send,',        [_('Qualifier'), 'audit deny', _('Access mode'), 'send',         _('Bus'), _('ALL'),  _('Path'), _('ALL'), _('Name'), _('ALL'),    _('Interface'), _('ALL'),  _('Member'), _('ALL'), _('Peer name'), _('ALL'),    _('Peer label'), _('ALL')]),
        ('dbus bind name=bind.name,',    [                              _('Access mode'), 'bind',         _('Bus'), _('ALL'),  _('Path'), _('ALL'), _('Name'), 'bind.name', _('Interface'), _('ALL'),  _('Member'), _('ALL'), _('Peer name'), _('ALL'),    _('Peer label'), _('ALL')]),  # noqa: E201
        ('dbus send bus=session path=/path interface=aa.test member=ExMbr peer=(name=(peer.name)),',
                                         [                              _('Access mode'), 'send',         _('Bus'), 'session', _('Path'), '/path',  _('Name'), _('ALL'),    _('Interface'), 'aa.test', _('Member'), 'ExMbr',  _('Peer name'), 'peer.name', _('Peer label'), _('ALL')]),  # noqa: E201,E127
        ('dbus send peer=(label=foo),',  [                              _('Access mode'), 'send',         _('Bus'), _('ALL'),  _('Path'), _('ALL'), _('Name'), _('ALL'),    _('Interface'), _('ALL'),  _('Member'), _('ALL'), _('Peer name'), _('ALL'),    _('Peer label'), 'foo']),  # noqa: E201
    )

    def _run_test(self, params, expected):
        obj = DbusRule.create_instance(params)
        self.assertEqual(obj.logprof_header(), expected)


# --- tests for DbusRuleset --- #

class DbusRulesTest(AATest):
    def test_empty_ruleset(self):
        ruleset = DbusRuleset()
        ruleset_2 = DbusRuleset()
        self.assertEqual([], ruleset.get_raw(2))
        self.assertEqual([], ruleset.get_clean(2))
        self.assertEqual([], ruleset_2.get_raw(2))
        self.assertEqual([], ruleset_2.get_clean(2))

    def test_ruleset_1(self):
        ruleset = DbusRuleset()
        rules = [
            'dbus peer=(label=foo),',
            'dbus send,',
        ]

        expected_raw = [
            'dbus peer=(label=foo),',
            'dbus send,',
            '',
        ]

        expected_clean = [
            'dbus peer=(label=foo),',
            'dbus send,',
            '',
        ]

        for rule in rules:
            ruleset.add(DbusRule.create_instance(rule))

        self.assertEqual(expected_raw, ruleset.get_raw())
        self.assertEqual(expected_clean, ruleset.get_clean())

    def test_ruleset_2(self):
        ruleset = DbusRuleset()
        rules = [
            'dbus send peer=(label=foo),',
            'allow dbus send,',
            'deny dbus bus=session, # example comment',
        ]

        expected_raw = [
            '  dbus send peer=(label=foo),',
            '  allow dbus send,',
            '  deny dbus bus=session, # example comment',
            '',
        ]

        expected_clean = [
            '  deny dbus bus=session, # example comment',
            '',
            '  allow dbus send,',
            '  dbus send peer=(label=foo),',
            '',
        ]

        for rule in rules:
            ruleset.add(DbusRule.create_instance(rule))

        self.assertEqual(expected_raw, ruleset.get_raw(1))
        self.assertEqual(expected_clean, ruleset.get_clean(1))


class DbusGlobTest(AATest):
    def setUp(self):
        self.maxDiff = None
        self.ruleset = DbusRuleset()

    def test_glob_1(self):
        self.assertEqual(self.ruleset.get_glob('dbus send,'), 'dbus,')

    # not supported or used yet
    # def test_glob_2(self):
    #     self.assertEqual(self.ruleset.get_glob('dbus send raw,'), 'dbus send,')

    def test_glob_ext(self):
        with self.assertRaises(NotImplementedError):
            # get_glob_ext is not available for dbus rules
            self.ruleset.get_glob_ext('dbus send peer=(label=foo),')


# class DbusDeleteTest(AATest):
#     pass


setup_all_loops(__name__)
if __name__ == '__main__':
    unittest.main(verbosity=1)



#!/usr/bin/python3
# ----------------------------------------------------------------------
#    Copyright (C) 2020 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import os
import shutil
import unittest
from collections import namedtuple

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.rule.include import IncludeRule, IncludeRuleset
from apparmor.translations import init_translation
from common_test import AATest, setup_all_loops, write_file

_ = init_translation()

exp = namedtuple(
    'exp', (  # 'audit', 'allow_keyword', 'deny',
        'comment', 'path', 'ifexists', 'ismagic'))

# --- tests for single IncludeRule --- #


class IncludeTest(AATest):
    def _compare_obj(self, obj, expected):
        self.assertEqual(False, obj.allow_keyword)  # not supported in include rules, expected to be always False
        self.assertEqual(False, obj.audit)          # not supported in include rules, expected to be always False
        self.assertEqual(False, obj.deny)           # not supported in include rules, expected to be always False
        self.assertEqual(None, obj.priority)        # not supported in include rules, expected to be always None
        self.assertEqual(expected.comment, obj.comment)

        self.assertEqual(expected.path, obj.path)
        self.assertEqual(expected.ifexists, obj.ifexists)
        self.assertEqual(expected.ismagic, obj.ismagic)


class IncludeTestParse(IncludeTest):
    tests = (
        # IncludeRule object                                      comment       path                 if exists  ismagic
        # #include
        ('#include <abstractions/base>',                      exp('',           'abstractions/base', False,     True)),  # magic path
        ('#include <abstractions/base> # comment',            exp(' # comment', 'abstractions/base', False,     True)),
        ('#include<abstractions/base>#comment',               exp(' #comment',  'abstractions/base', False,     True)),
        ('   #include     <abstractions/base>  ',             exp('',           'abstractions/base', False,     True)),
        ('#include "/foo/bar"',                               exp('',           '/foo/bar',          False,     False)),  # absolute path
        ('#include "/foo/bar" # comment',                     exp(' # comment', '/foo/bar',          False,     False)),
        ('#include "/foo/bar"#comment',                       exp(' #comment',  '/foo/bar',          False,     False)),
        ('   #include "/foo/bar"  ',                          exp('',           '/foo/bar',          False,     False)),
        # include (without #)
        ('include <abstractions/base>',                       exp('',           'abstractions/base', False,     True)),  # magic path
        ('include <abstractions/base> # comment',             exp(' # comment', 'abstractions/base', False,     True)),
        ('include<abstractions/base>#comment',                exp(' #comment',  'abstractions/base', False,     True)),
        ('   include     <abstractions/base>  ',              exp('',           'abstractions/base', False,     True)),
        ('include "/foo/bar"',                                exp('',           '/foo/bar',          False,     False)),  # absolute path
        ('include "/foo/bar" # comment',                      exp(' # comment', '/foo/bar',          False,     False)),
        ('include "/foo/bar"#comment',                        exp(' #comment',  '/foo/bar',          False,     False)),
        ('   include "/foo/bar"  ',                           exp('',           '/foo/bar',          False,     False)),
        # #include if exists
        ('#include if exists <abstractions/base>',            exp('',           'abstractions/base', True,      True)),  # magic path
        ('#include if exists <abstractions/base> # comment',  exp(' # comment', 'abstractions/base', True,      True)),
        ('#include if exists<abstractions/base>#comment',     exp(' #comment',  'abstractions/base', True,      True)),
        ('   #include    if     exists<abstractions/base>  ', exp('',           'abstractions/base', True,      True)),
        ('#include if exists "/foo/bar"',                     exp('',           '/foo/bar',          True,      False)),  # absolute path
        ('#include if exists "/foo/bar" # comment',           exp(' # comment', '/foo/bar',          True,      False)),
        ('#include if exists "/foo/bar"#comment',             exp(' #comment',  '/foo/bar',          True,      False)),
        ('   #include if exists "/foo/bar"  ',                exp('',           '/foo/bar',          True,      False)),
        # include if exists (without #)
        ('include if exists <abstractions/base>',             exp('',           'abstractions/base', True,      True)),  # magic path
        ('include if exists <abstractions/base> # comment',   exp(' # comment', 'abstractions/base', True,      True)),
        ('include if exists<abstractions/base>#comment',      exp(' #comment',  'abstractions/base', True,      True)),
        ('   include    if     exists<abstractions/base>  ',  exp('',           'abstractions/base', True,      True)),
        ('include if exists "/foo/bar"',                      exp('',           '/foo/bar',          True,      False)),  # absolute path
        ('include if exists "/foo/bar" # comment',            exp(' # comment', '/foo/bar',          True,      False)),
        ('include if exists "/foo/bar"#comment',              exp(' #comment',  '/foo/bar',          True,      False)),
        ('   include if exists "/foo/bar"  ',                 exp('',           '/foo/bar',          True,      False)),
    )

    def _run_test(self, rawrule, expected):
        self.assertTrue(IncludeRule.match(rawrule))
        obj = IncludeRule.create_instance(rawrule)
        self.assertEqual(rawrule.strip(), obj.raw_rule)
        self._compare_obj(obj, expected)


class IncludeTestParseInvalid(IncludeTest):
    tests = (
        # (' some #include if exists <abstractions/base>', AppArmorException),
        # ('  /etc/fstab r,',                              AppArmorException),
        # ('/usr/include r,',                              AppArmorException),
        # ('/include r,',                                  AppArmorException),
    )

    def _run_test(self, rawrule, expected):
        self.assertTrue(IncludeRule.match(rawrule))  # the above invalid rules still match the main regex!
        with self.assertRaises(expected):
            IncludeRule.create_instance(rawrule)

# class IncludeTestParseFromLog(IncludeTest):  # we'll never have log events for includes


class IncludeFromInit(IncludeTest):
    tests = (
        # IncludeRule object            ifexists  ismagic                     comment  path               ifexists  ismagic
        (IncludeRule('abstractions/base', False, False),                  exp('',      'abstractions/base', False, False)),
        (IncludeRule('foo',               True,  False),                  exp('',      'foo',               True,  False)),
        (IncludeRule('bar',               False, True),                   exp('',      'bar',               False, True)),
        (IncludeRule('baz',               True,  True),                   exp('',      'baz',               True,  True)),
        (IncludeRule('comment',           False, False, comment='# cmt'), exp('# cmt', 'comment',           False, False)),
    )

    def _run_test(self, obj, expected):
        self._compare_obj(obj, expected)


class InvalidIncludeInit(AATest):
    tests = (
        # init params            expected exception
        ((False,  False, False), AppArmorBug),  # wrong type for path
        (('',     False, False), AppArmorBug),  # empty path
        ((None,   False, False), AppArmorBug),  # wrong type for path
        # (('    ', False, False), AppArmorBug),  # whitespace-only path
        (('foo',  None,  False), AppArmorBug),  # wrong type for ifexists
        (('foo',  '',    False), AppArmorBug),  # wrong type for ifexists
        (('foo',  False, None),  AppArmorBug),  # wrong type for ismagic
        (('foo',  False, ''),    AppArmorBug),  # wrong type for ismagic
    )

    def _run_test(self, params, expected):
        with self.assertRaises(expected):
            IncludeRule(*params)

    def test_missing_params_1(self):
        with self.assertRaises(TypeError):
            IncludeRule()

    def test_missing_params_2(self):
        with self.assertRaises(TypeError):
            IncludeRule('foo')

    def test_missing_params_3(self):
        with self.assertRaises(TypeError):
            IncludeRule('foo', False)

    def test_audit_true(self):
        with self.assertRaises(AppArmorBug):
            IncludeRule('foo', False, False, audit=True)

    def test_deny_true(self):
        with self.assertRaises(AppArmorBug):
            IncludeRule('foo', False, False, deny=True)

    def test_priority_true(self):
        with self.assertRaises(AppArmorBug):
            IncludeRule('foo', False, False, priority=0)


class InvalidIncludeTest(AATest):
    def _check_invalid_rawrule(self, rawrule, matches_regex=False):
        obj = None
        self.assertEqual(IncludeRule.match(rawrule), matches_regex)
        with self.assertRaises(AppArmorException):
            obj = IncludeRule.create_instance(rawrule)

        self.assertIsNone(obj, 'IncludeRule handed back an object unexpectedly')

    def test_invalid_include_missing_path(self):
        self._check_invalid_rawrule('include', matches_regex=True)  # missing path

    def test_invalid_non_IncludeRule(self):
        self._check_invalid_rawrule('dbus,')  # not a include rule

    # def test_empty_data_1(self):
    #     obj = IncludeRule('foo', False, False)
    #     obj.path = ''
    #     # no path set
    #     with self.assertRaises(AppArmorBug):
    #         obj.get_clean(1)


class WriteIncludeTestAATest(AATest):
    def _run_test(self, rawrule, expected):
        self.assertTrue(IncludeRule.match(rawrule))
        obj = IncludeRule.create_instance(rawrule)
        clean = obj.get_clean()
        raw = obj.get_raw()

        self.assertEqual(expected.strip(), clean, 'unexpected clean rule')
        self.assertEqual(rawrule.strip(), raw, 'unexpected raw rule')

    tests = (
        #  (raw rule, clean rule)
        ('     include      <foo>            ',             'include <foo>'),
        # ('     include       foo             ',             'include "foo"'),  # several test cases disabled due to implementation restrictions, see re_match_include_parse()
        # ('     include      "foo"            ',             'include "foo"'),
        # ('     include       /foo             ',            'include "/foo"'),
        ('     include      "/foo"            ',            'include "/foo"'),

        ('     include      <foo>  # bar     ',             'include <foo> # bar'),
        # ('     include       foo   # bar     ',             'include "foo" # bar'),
        # ('     include      "foo"  # bar     ',             'include "foo" # bar'),
        # ('     include       /foo   # bar     ',            'include "/foo" # bar'),
        ('     include      "/foo"  # bar     ',            'include "/foo" # bar'),

        ('     include if    exists     <foo>            ', 'include if exists <foo>'),
        # ('     include if    exists      foo             ', 'include if exists "foo"'),
        # ('     include if    exists     "foo"            ', 'include if exists "foo"'),
        # ('     include if    exists      /foo            ', 'include if exists "/foo"'),
        ('     include if    exists     "/foo"           ', 'include if exists "/foo"'),

        # and the same again with #include...
        ('    #include      <foo>            ',             'include <foo>'),
        # ('    #include       foo             ',             'include "foo"'),
        # ('    #include      "foo"            ',             'include "foo"'),
        # ('    #include       /foo             ',            'include "/foo"'),
        ('    #include      "/foo"            ',            'include "/foo"'),

        ('    #include      <foo>  # bar     ',             'include <foo> # bar'),
        # ('    #include       foo   # bar     ',             'include "foo" # bar'),
        # ('    #include      "foo"  # bar     ',             'include "foo" # bar'),
        # ('    #include       /foo   # bar     ',            'include "/foo" # bar'),
        ('    #include      "/foo"  # bar     ',            'include "/foo" # bar'),

        ('    #include if    exists     <foo>            ', 'include if exists <foo>'),
        # ('    #include if    exists      foo             ', 'include if exists "foo"'),
        # ('    #include if    exists     "foo"            ', 'include if exists "foo"'),
        # ('    #include if    exists      /foo            ', 'include if exists "/foo"'),
        ('    #include if    exists     "/foo"           ', 'include if exists "/foo"'),
    )

    def test_write_manually(self):
        obj = IncludeRule('abs/foo', False, True, comment=' # cmt')

        expected = '    include <abs/foo> # cmt'

        self.assertEqual(expected, obj.get_clean(2), 'unexpected clean rule')
        self.assertEqual(expected, obj.get_raw(2), 'unexpected raw rule')


class IncludeCoveredTest(AATest):
    def _run_test(self, param, expected):
        obj = IncludeRule.create_instance(self.rule)
        check_obj = IncludeRule.create_instance(param)

        self.assertTrue(IncludeRule.match(param))

        self.assertEqual(obj.is_equal(check_obj), expected[0], 'Mismatch in is_equal, expected {}'.format(expected[0]))
        self.assertEqual(obj.is_equal(check_obj, True), expected[1], 'Mismatch in is_equal/strict, expected {}'.format(expected[1]))

        self.assertEqual(obj.is_covered(check_obj), expected[2], 'Mismatch in is_covered, expected {}'.format(expected[2]))
        self.assertEqual(obj.is_covered(check_obj, True, True), expected[3], 'Mismatch in is_covered/exact, expected {}'.format(expected[3]))


class IncludeCoveredTest_01(IncludeCoveredTest):
    rule = 'include <foo>'

    tests = (
        #   rule                      equal  strict equal  covered  covered exact
        ('include <foo>',            (True,  True,         True,    True)),
        ('#include <foo>',           (True,  False,        True,    True)),
        ('include if exists <foo>',  (False, False,        True,    True)),
        ('#include if exists <foo>', (False, False,        True,    True)),
        ('include <foobar>',         (False, False,        False,   False)),
        # ('include "foo"',            (False, False,        False,   False)),  # disabled due to implementation restrictions, see re_match_include_parse()
        # ('include if exists "foo"',  (False, False,        False,   False)),
    )


class IncludeCoveredTest_02(IncludeCoveredTest):
    rule = 'include if exists <foo>'

    tests = (
        #   rule                      equal  strict equal  covered  covered exact
        ('include <foo>',            (False, False,        False,   False)),
        ('#include <foo>',           (False, False,        False,   False)),
        ('#include if exists <foo>', (True,  False,        True,    True)),
        ('include <foobar>',         (False, False,        False,   False)),
        # ('include "foo"',            (False, False,        False,   False)),  # disabled due to implementation restrictions, see re_match_include_parse()
        # ('include if exists "foo"',  (False, False,        False,   False)),
    )


class IncludeCoveredTest_Invalid(AATest):
    # def test_borked_obj_is_covered_1(self):
    #     obj = IncludeRule.create_instance('include <foo>')
    #
    #     testobj = IncludeRule('foo', True, True)
    #     testobj.path = ''
    #
    #     with self.assertRaises(AppArmorBug):
    #         obj.is_covered(testobj)

    def test_invalid_is_covered(self):
        raw_rule = 'include <abstractions/base>'

        class SomeOtherClass(IncludeRule):
            pass

        obj = IncludeRule.create_instance(raw_rule)
        testobj = SomeOtherClass.create_instance(raw_rule)  # different type
        with self.assertRaises(AppArmorBug):
            obj.is_covered(testobj)

    def test_invalid_is_equal(self):
        raw_rule = 'include <abstractions/base>'

        class SomeOtherClass(IncludeRule):
            pass

        obj = IncludeRule.create_instance(raw_rule)
        testobj = SomeOtherClass.create_instance(raw_rule)  # different type
        with self.assertRaises(AppArmorBug):
            obj.is_equal(testobj)


class IncludeLogprofHeaderTest(AATest):
    tests = (
        ('include <abstractions/base>', [_('Include'), 'include <abstractions/base>']),
        ('include "/what/ever"',        [_('Include'), 'include "/what/ever"']),
    )

    def _run_test(self, params, expected):
        obj = IncludeRule.create_instance(params)
        self.assertEqual(obj.logprof_header(), expected)


class IncludeFullPathsTest(AATest):
    def AASetup(self):
        self.createTmpdir()

        # copy the local profiles to the test directory
        self.profile_dir = self.tmpdir + '/profiles'
        shutil.copytree('../../profiles/apparmor.d/', self.profile_dir, symlinks=True)

        inc_dir = os.path.join(self.profile_dir, 'abstractions/inc.d')
        os.mkdir(inc_dir, 0o755)
        write_file(inc_dir, 'incfoo', '/incfoo r,')
        write_file(inc_dir, 'incbar', '/incbar r,')
        write_file(inc_dir, 'README', '# README')  # gets skipped

        sub_dir = os.path.join(self.profile_dir, 'abstractions/inc.d/subdir')  # gets skipped
        os.mkdir(sub_dir, 0o755)

        empty_dir = os.path.join(self.profile_dir, 'abstractions/empty.d')
        os.mkdir(empty_dir, 0o755)

    tests = (
        #                                                @@ will be replaced with self.profile_dir
        ('include <abstractions/base>',                ('@@/abstractions/base',)),
        # ('include "foo"',                              ('@@/foo',)),  # TODO: adjust logic to honor quoted vs. magic paths (and allow quoted relative paths in re_match_include_parse())
        ('include "/foo/bar"',                         ('/foo/bar',)),
        ('include <abstractions/inc.d>',               ('@@/abstractions/inc.d/incbar', '@@/abstractions/inc.d/incfoo')),
        ('include <abstractions/empty.d>',             ()),
        ('include <abstractions/not_found>',           ('@@/abstractions/not_found',)),
        ('include if exists <abstractions/not_found>', ()),
    )

    def _run_test(self, params, expected):
        exp2 = []
        for path in expected:
            exp2.append(path.replace('@@', self.profile_dir))

        obj = IncludeRule.create_instance(params)
        self.assertEqual(obj.get_full_paths(self.profile_dir), exp2)


# --- tests for IncludeRuleset --- #

class IncludeRulesTest(AATest):
    def AASetup(self):
        self.createTmpdir()

        # copy the local profiles to the test directory
        self.profile_dir = self.tmpdir + '/profiles'
        shutil.copytree('../../profiles/apparmor.d/', self.profile_dir, symlinks=True)

        write_file(self.profile_dir, 'baz', '/baz r,')

    def test_empty_ruleset(self):
        ruleset = IncludeRuleset()
        ruleset_2 = IncludeRuleset()
        self.assertEqual([], ruleset.get_raw(2))
        self.assertEqual([], ruleset.get_clean(2))
        self.assertEqual([], ruleset_2.get_raw(2))
        self.assertEqual([], ruleset_2.get_clean(2))
        self.assertEqual([], ruleset_2.get_clean_unsorted(2))
        self.assertEqual([], ruleset.get_all_full_paths(self.profile_dir))

    def test_ruleset_1(self):
        ruleset = IncludeRuleset()
        rules = (
            ' include  <foo>  ',
            ' #include   "/bar" ',
        )

        expected_raw = [
            'include  <foo>',
            '#include   "/bar"',
            '',
        ]

        expected_clean = [
            'include "/bar"',
            'include <foo>',
            '',
        ]

        expected_clean_unsorted = [
            'include <foo>',
            'include "/bar"',
            '',
        ]

        expected_fullpaths = [
            os.path.join(self.profile_dir, 'foo'),
            '/bar'
        ]

        for rule in rules:
            ruleset.add(IncludeRule.create_instance(rule))

        self.assertEqual(expected_raw, ruleset.get_raw())
        self.assertEqual(expected_clean, ruleset.get_clean())
        self.assertEqual(expected_clean_unsorted, ruleset.get_clean_unsorted())
        self.assertEqual(expected_fullpaths, ruleset.get_all_full_paths(self.profile_dir))

    def test_ruleset_2(self):
        ruleset = IncludeRuleset()
        rules = (
            '   include   if  exists  <baz> ',
            ' include  <foo>  ',
            ' #include   "/bar" ',
            '#include if   exists   "/asdf"  ',
        )

        expected_raw = [
            'include   if  exists  <baz>',
            'include  <foo>',
            '#include   "/bar"',
            '#include if   exists   "/asdf"',
            '',
        ]

        expected_clean = [
            'include "/bar"',
            'include <foo>',
            'include if exists "/asdf"',
            'include if exists <baz>',
            '',
        ]

        expected_clean_unsorted = [
            'include if exists <baz>',
            'include <foo>',
            'include "/bar"',
            'include if exists "/asdf"',
            '',
        ]

        expected_fullpaths = [
            os.path.join(self.profile_dir, 'baz'),
            os.path.join(self.profile_dir, 'foo'),
            '/bar',
        ]

        for rule in rules:
            ruleset.add(IncludeRule.create_instance(rule))

        self.assertEqual(expected_raw, ruleset.get_raw())
        self.assertEqual(expected_clean, ruleset.get_clean())
        self.assertEqual(expected_clean_unsorted, ruleset.get_clean_unsorted())
        self.assertEqual(expected_fullpaths, ruleset.get_all_full_paths(self.profile_dir))


class IncludeGlobTestAATest(AATest):
    def setUp(self):
        self.maxDiff = None
        self.ruleset = IncludeRuleset()

    # def test_glob(self):
    #     with self.assertRaises(NotImplementedError):
    #         # get_glob_ext is not available for include rules
    #         self.ruleset.get_glob('include send set=int,')

    def test_glob_ext(self):
        with self.assertRaises(NotImplementedError):
            # get_glob_ext is not available for include rules
            self.ruleset.get_glob_ext('include send set=int,')


# class IncludeDeleteTestAATest(AATest):
#     pass


setup_all_loops(__name__)
if __name__ == '__main__':
    unittest.main(verbosity=1)
